"""Class for RESQML Wellbore Interpretation organizational objects."""

from ._utils import (equivalent_extra_metadata, alias_for_attribute, extract_has_occurred_during,
                     equivalent_chrono_pairs, create_xml_has_occurred_during)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns
from .wellbore_feature import WellboreFeature


class WellboreInterpretation(BaseResqpy):
    """Class for RESQML Wellbore Interpretation organizational objects.

    RESQML documentation:

       May refer to one of these:

       * **Wellbore**. A unique, oriented path from the bottom of a drilled borehole to the surface of the earth.
         The path must not overlap or cross itself.
       * **Borehole**. A hole excavated in the earth as a result of drilling or boring operations. The borehole
         may represent the hole of an entire wellbore (when no sidetracks are present), or a sidetrack extension.
         A borehole extends from an originating point (the surface location for the initial borehole or kickoff point
         for sidetracks) to a terminating (bottomhole) point.
       * **Sidetrack**. A borehole that originates in another borehole as opposed to originating at the surface.
    """

    resqml_type = 'WellboreInterpretation'
    valid_domains = ('depth', 'time', 'mixed')

    def __init__(self,
                 parent_model,
                 root_node = None,
                 uuid = None,
                 title = None,
                 is_drilled = None,
                 wellbore_feature = None,
                 domain = 'depth',
                 extra_metadata = None):
        """Initialises a wellbore interpretation organisational object."""

        # note: will create a paired WellboreFeature object when loading from xml

        self.is_drilled = is_drilled
        self.wellbore_feature = wellbore_feature
        self.feature_root = None if self.wellbore_feature is None else self.wellbore_feature.root
        if (not title) and self.wellbore_feature is not None:
            title = self.wellbore_feature.feature_name
        self.domain = domain
        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         extra_metadata = extra_metadata,
                         root_node = root_node)

    def _load_from_xml(self):
        root_node = self.root
        self.is_drilled = rqet.find_tag_bool(root_node, 'IsDrilled')
        self.domain = rqet.find_tag_text(root_node, 'Domain')
        interp_feature_ref_node = rqet.find_tag(root_node, 'InterpretedFeature')
        if interp_feature_ref_node is not None:
            self.feature_root = self.model.referenced_node(interp_feature_ref_node)
            if self.feature_root is not None:
                self.wellbore_feature = WellboreFeature(self.model,
                                                        uuid = self.feature_root.attrib['uuid'],
                                                        feature_name = self.model.title_for_root(self.feature_root))

    def iter_trajectories(self):
        """Iterable of associated trajectories."""

        import resqpy.well

        uuids = self.model.uuids(obj_type = "WellboreTrajectoryRepresentation", related_uuid = self.uuid)
        for uuid in uuids:
            yield resqpy.well.Trajectory(self.model, uuid = uuid)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, WellboreInterpretation):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.wellbore_feature is not None:
            if not self.wellbore_feature.is_equivalent(other.wellbore_feature):
                return False
        elif other.wellbore_feature is not None:
            return False
        if self.root is not None and other.root is not None:
            if rqet.citation_title_for_node(self.root) != rqet.citation_title_for_node(other.root):
                return False
            if self.domain != other.domain:
                return False
        elif self.root is not None or other.root is not None:
            return False
        if check_extra_metadata and not equivalent_extra_metadata(self, other):
            return False
        return (self.title == other.title and self.is_drilled == other.is_drilled)

    def create_xml(self,
                   wellbore_feature_root = None,
                   add_as_part = True,
                   add_relationships = True,
                   originator = None,
                   title_suffix = None,
                   reuse = True):
        """Creates a wellbore interpretation organisational xml node from a wellbore interpretation object."""

        # note: related wellbore feature node should be created first and referenced here

        if not self.title:
            self.title = self.wellbore_feature.feature_name
        if title_suffix:
            self.title += ' ' + title_suffix

        if reuse and self.try_reuse():
            return self.root
        wi = super().create_xml(add_as_part = False, originator = originator)

        if self.wellbore_feature is not None:
            wbf_root = self.wellbore_feature.root
            if wbf_root is not None:
                if wellbore_feature_root is None:
                    wellbore_feature_root = wbf_root
                else:
                    assert wbf_root is wellbore_feature_root, 'wellbore feature mismatch'

        if self.is_drilled is None:
            self.is_drilled = False

        id_node = rqet.SubElement(wi, ns['resqml2'] + 'IsDrilled')
        id_node.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
        id_node.text = str(self.is_drilled).lower()

        assert self.domain in self.valid_domains, 'illegal domain value for wellbore interpretation'
        domain_node = rqet.SubElement(wi, ns['resqml2'] + 'Domain')
        domain_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        domain_node.text = str(self.domain).lower()

        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title_for_root(wellbore_feature_root),
                                   wellbore_feature_root.attrib['uuid'],
                                   content_type = 'obj_WellboreFeature',
                                   root = wi)

        if add_as_part:
            self.model.add_part('obj_WellboreInterpretation', self.uuid, wi)
            if add_relationships:
                self.model.create_reciprocal_relationship(wi, 'destinationObject', wellbore_feature_root,
                                                          'sourceObject')

        return wi



# resqpy stratigraphy test

import os

import numpy as np

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.organize as rqo
import resqpy.strata as strata
# import resqpy.organize as rqo_new

# the following creates a fully related set of objects, saves to persistent storage, and retrieves


def test_strata(tmp_path):

    epc = os.path.join(tmp_path, 'strata.epc')

    model = rq.new_model(epc_file = epc)

    # earth model organisational objects

    emf = rqo.OrganizationFeature(model, feature_name = 'strata model', organization_kind = 'earth model')
    emf.create_xml()

    emi = rqo.EarthModelInterpretation(model, organization_feature = emf)
    emi.create_xml()

    # stratigraphic organisational objects
    # note: 'intrusive mud' has a spurious trailing space in the RESQML 2.0.1 xsd; resqpy accepts with or without

    unit_list = ['deep unit', 'mudstone unit', 'middle unit', 'shallow unit']
    composition_list = ['sedimentary siliclastic', 'intrusive mud', 'carbonate', 'sedimentary turbidite']
    implacement_list = ['autochtonous', 'allochtonous', 'autochtonous', 'autochtonous']
    mode_list = ['parallel to bottom', None, 'proportional between top and bottom', 'parallel to top']
    suf_list = []
    sui_list = []

    for unit_name, composition, implacement, mode in zip(unit_list, composition_list, implacement_list, mode_list):
        suf = strata.StratigraphicUnitFeature(model, title = unit_name)
        suf.create_xml()
        suf_list.append(suf)
        min_thick, max_thick = (250.0, 300.0) if unit_name == 'mudstone unit' else (None, None)
        sui = strata.StratigraphicUnitInterpretation(model,
                                                     title = unit_name,
                                                     stratigraphic_unit_feature = suf,
                                                     composition = composition,
                                                     material_implacement = implacement,
                                                     deposition_mode = mode,
                                                     min_thickness = min_thick,
                                                     max_thickness = max_thick,
                                                     thickness_uom = 'cm')
        sui.create_xml()
        sui_list.append(sui)

    # horizon organisational objects

    hf_list = []
    hi_list = []

    for unit_name in unit_list:
        base_horizon_name = 'base ' + unit_name
        hf = rqo.GeneticBoundaryFeature(model, feature_name = base_horizon_name, kind = 'horizon')
        hf.create_xml()
        hf_list.append(hf)
        hi = rqo.HorizonInterpretation(model,
                                       title = base_horizon_name,
                                       genetic_boundary_feature = hf,
                                       sequence_stratigraphy_surface = 'maximum flooding')
        hi.create_xml()
        hi_list.append(hi)

    # the main stratigraphic column rank interpretation

    scri = strata.StratigraphicColumnRank(model,
                                          earth_model_feature_uuid = emf.uuid,
                                          strata_uuid_list = [sui.uuid for sui in sui_list],
                                          title = 'stratigraphy')
    scri.set_contacts_from_horizons(horizon_uuids = [hi.uuid for hi in hi_list[1:]],
                                    older_contact_mode = 'erosion',
                                    younger_contact_mode = 'baselap')
    scri.create_xml()

    # the all encompassing stratigraphic column (in case of multiple ranks)

    sc = strata.StratigraphicColumn(model, rank_uuid_list = [scri.uuid], title = scri.title)
    sc.create_xml()

    # a small grid to relate to the stratigraphy, with a few layers per strata but a K gap for the mudstone

    crs = rqc.Crs(model)
    crs.create_xml()

    grid = grr.RegularGrid(model,
                           extent_kji = (10, 2, 3),
                           dxyz = (50.0, 50.0, 3.0),
                           origin = (0.0, 0.0, 1000.0),
                           crs_uuid = crs.uuid,
                           set_points_cached = True,
                           find_properties = False,
                           title = 'tiny grid')

    # convert one layer into a K gap
    grid.nk -= 1
    grid.extent_kji = np.array((grid.nk, grid.nj, grid.ni), dtype = int)
    grid.k_gaps = 1
    grid.k_gap_after_array = np.array([False, False, False, False, False, True, False, False], dtype = bool)
    grid._set_k_raw_index_array()

    # set the stratigraphy mapping for the grid's layers (and K gap)
    # note that stratigraphic unit indices start deep (old) and increase with shallowness (young)
    grid.stratigraphic_column_rank_uuid = scri.uuid
    grid.stratigraphic_units = np.array([3, 3, 3, 3, 2, 2, 1, 0, 0, 0], dtype = int)

    # write the grid array data and create grid xml, including geometry
    # True is the default for the new stratigraphy argument anyway but included here for emphasis!
    grid.write_hdf5_from_caches(write_active = False, stratigraphy = True)
    grid.create_xml(write_geometry = True, write_active = False)

    # now write the epc

    model.store_epc()

    # re-open the model and check relationships are as they should be

    model = rq.Model(epc)
    assert model is not None
    grid = model.grid()
    assert grid is not None
    assert grid.stratigraphic_column_rank_uuid is not None
    assert grid.stratigraphic_units is not None
    assert len(grid.stratigraphic_units) == grid.nk_plus_k_gaps
    strata_column_uuid = model.uuid(obj_type = 'StratigraphicColumn')
    assert strata_column_uuid is not None and bu.matching_uuids(strata_column_uuid, sc.uuid)
    strata_column = strata.StratigraphicColumn(model, uuid = strata_column_uuid)
    assert strata_column is not None
    assert len(strata_column.ranks) == 1
    strata_column_ri_uuid = model.uuid(obj_type = 'StratigraphicColumnRankInterpretation',
                                       related_uuid = strata_column_uuid)
    assert strata_column_ri_uuid is not None and bu.matching_uuids(strata_column_ri_uuid, scri.uuid)
    assert bu.matching_uuids(strata_column_ri_uuid, strata_column.ranks[0].uuid)
    assert bu.matching_uuids(strata_column_ri_uuid, grid.stratigraphic_column_rank_uuid)
    earth_model_feature_uuid = model.uuid(obj_type = 'OrganizationFeature', related_uuid = strata_column_ri_uuid)
    assert earth_model_feature_uuid is not None
    earth_model_feature = rqo.OrganizationFeature(model, uuid = earth_model_feature_uuid)
    assert earth_model_feature.organization_kind == 'earth model'
    strata_column_ri = strata.StratigraphicColumnRank(model, uuid = strata_column_ri_uuid)
    assert strata_column_ri is not None
    sui_uuid_set = set(model.uuids(obj_type = 'StratigraphicUnitInterpretation', related_uuid = strata_column_ri_uuid))
    assert len(sui_uuid_set) == 4
    assert sui_uuid_set == set([sui.uuid for sui in strata_column_ri.iter_units()])
    hi_uuid_set = set(model.uuids(obj_type = 'HorizonInterpretation', related_uuid = strata_column_ri_uuid))
    assert len(hi_uuid_set) == 3
    pou_set = set()
    for contact in strata_column_ri.iter_contacts():
        if contact.part_of_uuid is not None:
            pou_set.add(contact.part_of_uuid)
    assert hi_uuid_set == pou_set

    # check column rank units and contacts are well ordered

    previous_unit_index = -1
    for i, sui_uuid in strata_column_ri.units:
        assert i > previous_unit_index
        previous_unit_index = i
        assert sui_uuid is not None

    previous_contact_index = -1
    for contact in strata_column_ri.contacts:
        assert contact.index > previous_contact_index
        previous_contact_index = contact.index

    # check that stratigraphy for grid layers is as expected

    # note: K gap is skipped over here
    expected_unit_names_per_layer = 4 * ['shallow unit'] + 2 * ['middle unit'] + 3 * ['deep unit']

    sui_name_list = [sui.title for sui in strata_column_ri.iter_units()]

    for k in range(grid.nk):
        unit = strata_column_ri.unit_for_unit_index(grid.stratigraphic_units[grid.k_raw_index_array[k]])
        assert unit is not None
        assert unit.title == expected_unit_names_per_layer[k]



import math as maths
import os

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_almost_equal

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.lines as rql
import resqpy.model as rq
import resqpy.olio.triangulation as triangulation
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.surface as rqs
import resqpy.unstructured as rug
from resqpy.olio.random_seed import seed


def test_hexa_grid_from_grid(example_model_with_properties):

    model = example_model_with_properties

    ijk_grid_uuid = model.uuid(obj_type = 'IjkGridRepresentation')
    assert ijk_grid_uuid is not None

    # create an unstructured grid with hexahedral cells from an unsplit IJK grid (includes write_hdf5 and create_xml)
    hexa = rug.HexaGrid.from_unsplit_grid(model, ijk_grid_uuid, inherit_properties = True, title = 'HEXA')

    hexa_uuid = hexa.uuid

    epc = model.epc_file
    assert epc

    model.store_epc()

    # re-open model and check hexa grid

    model = rq.Model(epc)

    unstructured_uuids = model.uuids(obj_type = 'UnstructuredGridRepresentation')

    assert unstructured_uuids is not None and len(unstructured_uuids) == 1

    hexa_grid = grr.any_grid(model, uuid = unstructured_uuids[0])
    assert isinstance(hexa_grid, rug.HexaGrid)

    assert hexa_grid.cell_shape == 'hexahedral'

    hexa_grid.check_indices()
    hexa_grid.check_hexahedral()

    # instantiate ijk grid and compare hexa grid with it
    ijk_grid = grr.any_grid(model, uuid = ijk_grid_uuid)
    assert ijk_grid is not None
    assert not np.any(np.isnan(ijk_grid.points_ref()))

    assert hexa_grid.cell_count == ijk_grid.cell_count()
    assert hexa_grid.active_cell_count() == hexa_grid.cell_count
    assert hexa_grid.node_count == (ijk_grid.nk + 1) * (ijk_grid.nj + 1) * (ijk_grid.ni + 1)
    assert hexa_grid.face_count == ((ijk_grid.nk + 1) * ijk_grid.nj * ijk_grid.ni + ijk_grid.nk *
                                    (ijk_grid.nj + 1) * ijk_grid.ni + ijk_grid.nk * ijk_grid.nj * (ijk_grid.ni + 1))

    assert bu.matching_uuids(hexa_grid.extract_crs_uuid(), ijk_grid.crs_uuid)
    assert hexa_grid.crs_is_right_handed == rqc.Crs(model, uuid = ijk_grid.crs_uuid).is_right_handed_xyz()

    # points arrays should be identical for the two grids
    assert not np.any(np.isnan(hexa_grid.points_ref()))
    assert_array_almost_equal(hexa_grid.points_ref(), ijk_grid.points_ref(masked = False).reshape((-1, 3)))

    # compare centre points of cells (not sure if these would be coincident for irregular shaped cells)
    # note that the centre_point() method exercises several other methods
    hexa_centres = hexa_grid.centre_point()
    ijk_centres = ijk_grid.centre_point()
    assert_array_almost_equal(hexa_centres, ijk_centres.reshape((-1, 3)), decimal = 3)

    # check that face and node indices are in range
    hexa_grid.cache_all_geometry_arrays()
    assert np.all(0 <= hexa_grid.faces_per_cell)
    assert np.all(hexa_grid.faces_per_cell < hexa_grid.face_count)
    assert np.all(0 <= hexa_grid.nodes_per_face)
    assert np.all(hexa_grid.nodes_per_face < hexa_grid.node_count)
    assert len(hexa_grid.faces_per_cell_cl) == hexa_grid.cell_count
    assert len(hexa_grid.nodes_per_face_cl) == hexa_grid.face_count

    # check distinct nodes for first cell
    cell_nodes = hexa_grid.distinct_node_indices_for_cell(0)  # is sorted by node index
    assert len(cell_nodes) == 8
    ni1_nj1 = (ijk_grid.ni + 1) * (ijk_grid.nj + 1)
    expected_nodes = np.array((0, 1, ijk_grid.ni + 1, ijk_grid.ni + 2, ni1_nj1, ni1_nj1 + 1, ni1_nj1 + ijk_grid.ni + 1,
                               ni1_nj1 + ijk_grid.ni + 2),
                              dtype = int)
    assert np.all(cell_nodes == expected_nodes)

    # check that some simple convenience methods work okay
    assert hexa_grid.face_count_for_cell(0) == 6
    assert hexa_grid.max_face_count_for_any_cell() == 6
    assert hexa_grid.max_node_count_for_any_face() == 4

    # check that correct number of edges is found for a face
    edges = hexa_grid.edges_for_face(hexa_grid.face_count // 2)  # arbitrary face in middle
    assert edges.shape == (4, 2)
    edges = hexa_grid.edges_for_face_with_node_indices_ordered_within_pairs(hexa_grid.face_count // 2)
    assert edges.shape == (4, 2)
    for a, b in edges:  # check node within pair ordering
        assert a < b

    # compare corner points for first cell with those for ijk grid cell
    cp = hexa_grid.corner_points(0)
    assert cp.shape == (8, 3)
    assert_array_almost_equal(cp.reshape((2, 2, 2, 3)),
                              ijk_grid.corner_points(cell_kji0 = (0, 0, 0), cache_resqml_array = False))

    # have a look at handedness of cell faces
    assert len(hexa_grid.cell_face_is_right_handed) == 6 * hexa_grid.cell_count
    # following assertion only applies to HexaGrid built from_unsplit_grid()
    assert np.count_nonzero(hexa_grid.cell_face_is_right_handed) == 3 * hexa_grid.cell_count  # half are right handed

    # compare cell volumes for first cell
    hexa_vol = hexa_grid.volume(0)
    ijk_vol = ijk_grid.volume(cell_kji0 = 0, cache_resqml_array = False, cache_volume_array = False)
    assert maths.isclose(hexa_vol, ijk_vol)
    assert maths.isclose(hexa_vol, 1.0, rel_tol = 1.0e-3)

    # check face normal for first face (K- face of first cell)
    assert_array_almost_equal(hexa_grid.face_normal(0), (0.0, 0.0, -1.0))

    # check that planar approximation of last face is about the same as the original
    fi = hexa_grid.face_count - 1
    assert_array_almost_equal(hexa_grid.planar_face_points(fi),
                              hexa_grid.points_cached[hexa_grid.node_indices_for_face(fi)],
                              decimal = 3)

    # construct a Delauney triangulation of the last face
    triangulated = hexa_grid.face_triangulation(fi, local_nodes = True)
    assert triangulated.shape == (2, 3)
    assert np.all(triangulated.flatten() < 4)
    assert np.all(np.unique(triangulated) == (0, 1, 2, 3))
    # also test the triangulation using the global node indices, for all the faces of the last cell
    for face_index in hexa_grid.face_indices_for_cell(hexa_grid.cell_count - 1):
        triangulated = hexa_grid.face_triangulation(face_index, local_nodes = False)
        assert triangulated.shape == (2, 3)

    # check the area of a middle face (the example model has unit area faces)
    assert maths.isclose(hexa_grid.area_of_face(hexa_grid.face_count // 2), 1.0, rel_tol = 1.0e-3)

    # compare properties
    ijk_pc = ijk_grid.extract_property_collection()
    hexa_pc = hexa_grid.extract_property_collection()
    assert ijk_pc is not None and hexa_pc is not None
    assert ijk_pc.number_of_parts() <= hexa_pc.number_of_parts()  # hexa grid has extra active array in this test
    for ijk_part in ijk_pc.parts():
        p_title = ijk_pc.citation_title_for_part(ijk_part)
        hexa_part = hexa_pc.singleton(citation_title = p_title)
        assert hexa_part is not None
        assert hexa_part != ijk_part  # properties are different objects with distinct uuids
        assert hexa_pc.continuous_for_part(hexa_part) == ijk_pc.continuous_for_part(ijk_part)
        ijk_array = ijk_pc.cached_part_array_ref(ijk_part)
        hexa_array = hexa_pc.cached_part_array_ref(hexa_part)
        assert ijk_array is not None and hexa_array is not None
        if hexa_pc.continuous_for_part(hexa_part):
            assert_array_almost_equal(hexa_array.flatten(), ijk_array.flatten())
        else:
            assert np.all(hexa_array.flatten() == ijk_array.flatten())

    # test TetraGrid.from_unstructured_cell is as expected for a hexahedral cell
    tetra = rug.TetraGrid.from_unstructured_cell(hexa_grid, hexa_grid.cell_count // 2)
    tetra.check_indices()
    tetra.check_tetra()
    assert tetra.cell_count == 12  # 2 tetrahedra per hexa face
    assert tetra.node_count == 9  # 8 nodes of hexa cell plus centre point
    assert tetra.face_count == 6 * 2 + 12 + 6
    # check total volume of tetra grid version of cell
    assert maths.isclose(tetra.grid_volume(), 1.0, rel_tol = 1.0e-3)


def test_tetra_grid(tmp_path):

    epc = os.path.join(tmp_path, 'tetra_test.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()

    # create an empty TetraGrid
    tetra = rug.TetraGrid(model, title = 'star')
    assert tetra.cell_shape == 'tetrahedral'

    # hand craft all attribute data
    tetra.crs_uuid = model.uuid(obj_type = 'LocalDepth3dCrs')
    assert tetra.crs_uuid is not None
    assert bu.matching_uuids(tetra.crs_uuid, crs.uuid)
    tetra.set_cell_count(5)
    # faces
    tetra.face_count = 16
    tetra.faces_per_cell_cl = np.arange(4, 4 * 5 + 1, 4, dtype = int)
    tetra.faces_per_cell = np.empty(20, dtype = int)
    tetra.faces_per_cell[:4] = (0, 1, 2, 3)  # cell 0
    tetra.faces_per_cell[4:8] = (0, 4, 5, 6)  # cell 1
    tetra.faces_per_cell[8:12] = (1, 7, 8, 9)  # cell 2
    tetra.faces_per_cell[12:16] = (2, 10, 11, 12)  # cell 3
    tetra.faces_per_cell[16:] = (3, 13, 14, 15)  # cell 4
    # nodes
    tetra.node_count = 8
    tetra.nodes_per_face_cl = np.arange(3, 3 * 16 + 1, 3, dtype = int)
    tetra.nodes_per_face = np.empty(48, dtype = int)
    # internal faces (cell 0)
    tetra.nodes_per_face[:3] = (0, 1, 2)  # face 0
    tetra.nodes_per_face[3:6] = (0, 3, 1)  # face 1
    tetra.nodes_per_face[6:9] = (1, 3, 2)  # face 2
    tetra.nodes_per_face[9:12] = (2, 3, 0)  # face 3
    # external faces (cell 1)
    tetra.nodes_per_face[12:15] = (0, 1, 4)  # face 4
    tetra.nodes_per_face[15:18] = (1, 2, 4)  # face 5
    tetra.nodes_per_face[18:21] = (2, 0, 4)  # face 6
    # external faces (cell 2)
    tetra.nodes_per_face[21:24] = (0, 3, 5)  # face 7
    tetra.nodes_per_face[24:27] = (3, 1, 5)  # face 8
    tetra.nodes_per_face[27:30] = (1, 0, 5)  # face 9
    # external faces (cell 3)
    tetra.nodes_per_face[30:33] = (1, 3, 6)  # face 10
    tetra.nodes_per_face[33:36] = (3, 2, 6)  # face 11
    tetra.nodes_per_face[36:39] = (2, 1, 6)  # face 12
    # external faces (cell 4)
    tetra.nodes_per_face[39:42] = (2, 3, 7)  # face 10
    tetra.nodes_per_face[42:45] = (3, 0, 7)  # face 11
    tetra.nodes_per_face[45:] = (0, 2, 7)  # face 12
    # face handedness
    tetra.cell_face_is_right_handed = np.zeros(20, dtype = bool)  # False for all faces for external cells (1 to 4)
    tetra.cell_face_is_right_handed[:4] = True  # True for all faces of internal cell (0)
    # points
    tetra.points_cached = np.zeros((8, 3))
    # internal cell (0) points
    half_edge = 36.152
    one_over_root_two = 1.0 / maths.sqrt(2.0)
    tetra.points_cached[0] = (-half_edge, 0.0, -half_edge * one_over_root_two)
    tetra.points_cached[1] = (half_edge, 0.0, -half_edge * one_over_root_two)
    tetra.points_cached[2] = (0.0, half_edge, half_edge * one_over_root_two)
    tetra.points_cached[3] = (0.0, -half_edge, half_edge * one_over_root_two)
    # project remaining nodes outwards
    for fi, o_node in enumerate((3, 2, 0, 1)):
        fc = tetra.face_centre_point(fi)
        tetra.points_cached[4 + fi] = fc - (tetra.points_cached[o_node] - fc)

    # basic validity check
    tetra.check_tetra()

    # write arrays, create xml and store model
    tetra.write_hdf5()
    tetra.create_xml()
    model.store_epc()

    # re-open model and establish grid
    model = rq.Model(epc)
    assert model is not None
    tetra_uuid = model.uuid(obj_type = 'UnstructuredGridRepresentation', title = 'star')
    assert tetra_uuid is not None
    tetra = rug.TetraGrid(model, uuid = tetra_uuid)
    assert tetra is not None
    # perform basic checks
    assert tetra.cell_count == 5
    assert tetra.cell_shape == 'tetrahedral'
    tetra.check_tetra()

    # test volume calculation
    expected_cell_volume = ((2.0 * half_edge)**3) / (6.0 * maths.sqrt(2.0))
    for cell in range(tetra.cell_count):
        assert maths.isclose(tetra.volume(cell), expected_cell_volume, rel_tol = 1.0e-3)
    assert maths.isclose(tetra.grid_volume(), 5.0 * expected_cell_volume)

    # test face area
    expected_area = maths.sqrt(3.0 * half_edge * (half_edge**3))
    area = tetra.area_of_face(0)
    assert maths.isclose(area, expected_area, rel_tol = 1.0e-3)

    # test internal / external face lists
    assert np.all(tetra.external_face_indices() == np.arange(4, 16, dtype = int))
    inactive_mask = np.zeros(5, dtype = bool)
    assert np.all(tetra.external_face_indices_for_masked_cells(inactive_mask) == tetra.external_face_indices())
    assert np.all(tetra.internal_face_indices_for_masked_cells(inactive_mask) == np.arange(4, dtype = int))
    # mask out central cell
    inactive_mask[0] = True
    assert len(tetra.external_face_indices_for_masked_cells(inactive_mask)) == tetra.face_count
    assert len(tetra.internal_face_indices_for_masked_cells(inactive_mask)) == 0


def test_vertical_prism_grid_from_surfaces(tmp_path):

    epc = os.path.join(tmp_path, 'vertical_prism.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()

    # create a point set representing a pentagon with a centre node
    pentagon_points = np.array([[-100.0, -200.0, 1050.0], [-200.0, 0.0, 1050.0], [0.0, 200.0, 1025.0],
                                [200.0, 0.0, 975.0], [100.0, -200.0, 999.0], [0.0, 0.0, 1000.0]])
    pentagon = rqs.PointSet(model, points_array = pentagon_points, crs_uuid = crs.uuid, title = 'pentagon')
    pentagon.write_hdf5()
    pentagon.create_xml()

    # create a surface from the point set (will make a Delauney triangulation)
    top_surf = rqs.Surface(model, point_set = pentagon, title = 'top surface')
    top_surf.write_hdf5()
    top_surf.create_xml()
    surf_list = [top_surf]

    # check the pentagon surface
    pentagon_triangles, pentagon_points = top_surf.triangles_and_points()
    assert pentagon_points.shape == (6, 3)
    assert pentagon_triangles.shape == (5, 3)

    # create a couple of horizontal surfaces at greater depths
    boundary = np.array([[-300.0, -300.0, 0.0], [300.0, 300.0, 0.0]])
    for depth in (1100.0, 1200.0):
        base = rqs.Surface(model)
        base.set_to_horizontal_plane(depth, boundary)
        base.write_hdf5()
        base.create_xml()
        surf_list.append(base)

    # now build a vertical prism grid from the surfaces
    grid = rug.VerticalPrismGrid.from_surfaces(model, surf_list, title = 'the pentagon')
    grid.write_hdf5()
    grid.create_xml()

    model.store_epc()

    # re-open model

    model = rq.Model(epc)
    assert model is not None

    # find grid by title
    grid_uuid = model.uuid(obj_type = 'UnstructuredGridRepresentation', title = 'the pentagon')
    assert grid_uuid is not None

    # re-instantiate the grid
    grid = rug.VerticalPrismGrid(model, uuid = grid_uuid)
    assert grid is not None
    assert grid.nk == 2
    assert grid.cell_count == 10
    assert grid.node_count == 18
    assert grid.face_count == 35

    # create a very similar grid using explicit triangulation arguments

    # make the same Delauney triangulation
    triangles = triangulation.dt(pentagon_points)

    # slightly shrink pentagon points to be within area of surfaces
    for i in range(len(pentagon_points)):
        if pentagon_points[i, 0] < 0.0:
            pentagon_points[i, 0] += 1.0
        elif pentagon_points[i, 0] > 0.0:
            pentagon_points[i, 0] -= 1.0
        if pentagon_points[i, 1] < 0.0:
            pentagon_points[i, 1] += 1.0
        elif pentagon_points[i, 1] > 0.0:
            pentagon_points[i, 1] -= 1.0

    # load the surfaces
    surf_uuids = model.uuids(obj_type = 'TriangulatedSetRepresentation', sort_by = 'oldest')
    surf_list = []
    for surf_uuid in surf_uuids:
        surf_list.append(rqs.Surface(model, uuid = surf_uuid))

    # create a new vertical prism grid using the explicit triangulation arguments
    similar = rug.VerticalPrismGrid.from_surfaces(model,
                                                  surf_list,
                                                  column_points = pentagon_points,
                                                  column_triangles = triangles,
                                                  title = 'similar pentagon')

    # check similarity
    for attr in ('cell_shape', 'nk', 'cell_count', 'node_count', 'face_count'):
        assert getattr(grid, attr) == getattr(similar, attr)
    for index_attr in ('nodes_per_face', 'nodes_per_face_cl', 'faces_per_cell', 'faces_per_cell_cl'):
        assert np.all(getattr(grid, index_attr) == getattr(similar, index_attr))
    assert_allclose(grid.points_ref(), similar.points_ref(), atol = 2.0)

    # check that isotropic horizontal permeability is preserved
    permeability = 250.0
    primary_k = np.full((grid.cell_count,), permeability)
    orthogonal_k = primary_k.copy()
    triple_k = grid.triple_horizontal_permeability(primary_k, orthogonal_k, 37.0)
    assert triple_k.shape == (grid.cell_count, 3)
    assert_array_almost_equal(triple_k, permeability)
    azimuth = np.linspace(0.0, 360.0, num = grid.cell_count)
    triple_k = grid.triple_horizontal_permeability(primary_k, orthogonal_k, azimuth)
    assert triple_k.shape == (grid.cell_count, 3)
    assert_array_almost_equal(triple_k, permeability)

    # check that anisotropic horizontal permeability is correctly bounded
    orthogonal_k *= 0.1
    triple_k = grid.triple_horizontal_permeability(primary_k, orthogonal_k, azimuth)
    assert triple_k.shape == (grid.cell_count, 3)
    assert np.all(triple_k <= permeability)
    assert np.all(triple_k >= permeability * 0.1)
    assert np.min(triple_k) < permeability / 2.0
    assert np.max(triple_k) > permeability / 2.0

    # set up some properties
    pc = grid.property_collection
    assert pc is not None
    pc.add_cached_array_to_imported_list(cached_array = None,
                                         source_info = 'unit test',
                                         keyword = 'NETGRS',
                                         property_kind = 'net to gross ratio',
                                         discrete = False,
                                         uom = 'm3/m3',
                                         indexable_element = 'cells',
                                         const_value = 0.75)
    pc.add_cached_array_to_imported_list(cached_array = None,
                                         source_info = 'unit test',
                                         keyword = 'PERMK',
                                         property_kind = 'permeability rock',
                                         facet_type = 'direction',
                                         facet = 'K',
                                         discrete = False,
                                         uom = 'mD',
                                         indexable_element = 'cells',
                                         const_value = 10.0)
    pc.add_cached_array_to_imported_list(cached_array = None,
                                         source_info = 'unit test',
                                         keyword = 'PERM',
                                         property_kind = 'permeability rock',
                                         facet_type = 'direction',
                                         facet = 'primary',
                                         discrete = False,
                                         uom = 'mD',
                                         indexable_element = 'cells',
                                         const_value = 100.0)
    pc.add_cached_array_to_imported_list(cached_array = None,
                                         source_info = 'unit test',
                                         keyword = 'PERM',
                                         property_kind = 'permeability rock',
                                         facet_type = 'direction',
                                         facet = 'orthogonal',
                                         discrete = False,
                                         uom = 'mD',
                                         indexable_element = 'cells',
                                         const_value = 20.0)
    x_min, x_max = grid.xyz_box()[:, 0]
    relative_x = (grid.centre_point()[:, 0] - x_min) * (x_max - x_min)
    azi = relative_x * 90.0 + 45.0
    pc.add_cached_array_to_imported_list(cached_array = azi,
                                         source_info = 'unit test',
                                         keyword = 'primary permeability azimuth',
                                         property_kind = 'plane angle',
                                         facet_type = 'direction',
                                         facet = 'primary',
                                         discrete = False,
                                         uom = 'dega',
                                         indexable_element = 'cells')
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model()

    model.store_epc()

    # test that half cell transmissibilities can be computed
    half_t = grid.half_cell_transmissibility()
    assert np.all(half_t > 0.0)

    # add the half cell transmissibility array as a property
    pc.add_cached_array_to_imported_list(cached_array = half_t.flatten(),
                                         source_info = 'unit test',
                                         keyword = 'half transmissibility',
                                         property_kind = 'transmissibility',
                                         discrete = False,
                                         count = 1,
                                         indexable_element = 'faces per cell')
    pc.write_hdf5_for_imported_list()
    pc.create_xml_for_imported_list_and_add_parts_to_model(extra_metadata = {'uom': 'm3.cP/(d.kPa)'})

    model.store_epc()


def test_vertical_prism_grid_from_seed_points_and_surfaces(tmp_path):

    seed(23487656)  # to ensure test reproducibility

    epc = os.path.join(tmp_path, 'voronoi_prism_grid.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()

    # define a boundary polyline:
    b_count = 7
    boundary_points = np.empty((b_count, 3))
    radius = 1000.0
    for i in range(b_count):
        theta = -vec.radians_from_degrees(i * 360.0 / b_count)
        boundary_points[i] = (2.0 * radius * maths.cos(theta), radius * maths.sin(theta), 0.0)
    boundary = rql.Polyline(model,
                            set_coord = boundary_points,
                            set_bool = True,
                            set_crs = crs.uuid,
                            title = 'rough ellipse')
    boundary.write_hdf5()
    boundary.create_xml()

    # derive a larger area of interest
    aoi = rql.Polyline.from_scaled_polyline(boundary, 1.1, title = 'area of interest')
    aoi.write_hdf5()
    aoi.create_xml()
    min_xy = np.min(aoi.coordinates[:, :2], axis = 0) - 50.0
    max_xy = np.max(aoi.coordinates[:, :2], axis = 0) + 50.0

    print(f'***** min max xy aoi+ : {min_xy} {max_xy}')  # debug

    # create some seed points within boundary
    seed_count = 5
    seeds = rqs.PointSet(model,
                         crs_uuid = crs.uuid,
                         polyline = boundary,
                         random_point_count = seed_count,
                         title = 'seeds')
    seeds.write_hdf5()
    seeds.create_xml()
    seeds_xy = seeds.single_patch_array_ref(0)

    for seed_xy in seeds_xy:
        assert aoi.point_is_inside_xy(seed_xy), f'seed point {seed_xy} outwith aoi'

    print(f'***** min max xy seeds : {np.min(seeds_xy, axis = 0)} {np.max(seeds_xy, axis = 0)}')  # debug

    # create some horizon surfaces
    ni, nj = 21, 11
    lattice = rqs.Mesh(model,
                       crs_uuid = crs.uuid,
                       mesh_flavour = 'regular',
                       ni = ni,
                       nj = nj,
                       origin = (min_xy[0], min_xy[1], 0.0),
                       dxyz_dij = np.array([[(max_xy[0] - min_xy[0]) / (ni - 1), 0.0, 0.0],
                                            [0.0, (max_xy[1] - min_xy[1]) / (nj - 1), 0.0]]))
    lattice.write_hdf5()
    lattice.create_xml()
    horizons = []
    for i in range(4):
        horizon_depths = 1000.0 + 100.0 * i + 20.0 * (np.random.random((nj, ni)) - 0.5)
        horizon_mesh = rqs.Mesh(model,
                                crs_uuid = crs.uuid,
                                mesh_flavour = 'ref&z',
                                ni = ni,
                                nj = nj,
                                z_values = horizon_depths,
                                z_supporting_mesh_uuid = lattice.uuid,
                                title = 'h' + str(i))
        horizon_mesh.write_hdf5()
        horizon_mesh.create_xml()
        horizon_surface = rqs.Surface(model,
                                      crs_uuid = crs.uuid,
                                      mesh = horizon_mesh,
                                      quad_triangles = True,
                                      title = horizon_mesh.title)
        horizon_surface.write_hdf5()
        horizon_surface.create_xml()
        horizons.append(horizon_surface)

    # create a re-triangulated Voronoi vertical prism grid
    grid = rug.VerticalPrismGrid.from_seed_points_and_surfaces(model,
                                                               seeds_xy,
                                                               horizons,
                                                               aoi,
                                                               title = "giant's causeway")
    assert grid is not None
    grid.write_hdf5()
    grid.create_xml()

    # check cell thicknesses are in expected range
    thick = grid.thickness()
    assert np.all(thick >= 80.0)
    assert np.all(thick <= 120.0)

    model.store_epc()



import pytest
import datetime as dt

from resqpy.time_series import TimeDuration


@pytest.mark.parametrize('days, hours, earlier_timestamp, later_timestamp',
                         [(1, 12, '2021-12-13T00:00:00Z', '2021-12-14T12:00:00Z'),
                          (30, 0, '2021-12-13T00:00:00Z', '2022-01-12T00:00:00Z'),
                          (30, 23, '2022-04-13T06:30:00Z', '2022-05-14T05:30:00Z'),
                          (365, 0, '2024-01-01T00:00:00Z', '2024-12-31T00:00:00Z'),
                          (0, 0, '2014-01-01T00:00:00Z', '2014-01-01T00:00:00Z')])
def test_timestamp_before_and_after_duration(days, hours, earlier_timestamp, later_timestamp):
    duration = TimeDuration(days = days, hours = hours)
    assert duration.timestamp_before_duration(later_timestamp) == earlier_timestamp
    assert duration.timestamp_after_duration(earlier_timestamp) == later_timestamp



"""Time series classes and functions."""

__all__ = [
    'AnyTimeSeries', 'GeologicTimeSeries', 'TimeDuration', 'TimeSeries', 'selected_time_series', 'simplified_timestamp',
    'cleaned_timestamp', 'time_series_from_list', 'merge_timeseries_from_uuid', 'geologic_time_str',
    'timeframe_for_time_series_uuid', 'any_time_series', 'time_series_from_nexus_summary'
]

from ._any_time_series import AnyTimeSeries
from ._geologic_time_series import GeologicTimeSeries
from ._time_duration import TimeDuration
from ._time_series import TimeSeries

from ._functions import selected_time_series, simplified_timestamp, cleaned_timestamp, time_series_from_list,  \
    merge_timeseries_from_uuid, geologic_time_str, timeframe_for_time_series_uuid, any_time_series
from ._from_nexus_summary import time_series_from_nexus_summary



import os
import numpy as np
import pandas as pd
import pytest
import math

from resqpy.grid import RegularGrid
from resqpy.model import Model
import resqpy.olio.vector_utilities as vec
import resqpy.well


def test_Trajectory_load_from_xml(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    epc_path = model.epc_file
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    datum.create_xml()
    mds = np.array([300, 310, 330, 340])
    zs = mds - elevation
    well_name = 'JubJub'
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [100, 120, 140, 160],
        'Y': [345, 365, 386, 400],
        'Z': zs,
        'WELL': ['JubJub', 'JubJub', 'JubJub', 'JubJub']
    })

    # Create a trajectory from dataframe
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')

    # --------- Act ----------
    # Save to disk
    trajectory.write_hdf5()
    trajectory.create_xml()
    trajectory_uuid = trajectory.uuid
    model.store_epc()
    model.h5_release()

    # Reload from disk
    model2 = Model(epc_file = epc_path)
    trajectory2 = resqpy.well.Trajectory(model2, uuid = trajectory_uuid, set_tangent_vectors = True)

    # --------- Assert --------------
    # Check all attributes were loaded from disk correctly
    assert trajectory2 is not None
    assert trajectory2.well_name == well_name
    np.testing.assert_almost_equal(trajectory2.measured_depths, mds)
    assert trajectory2.md_datum == datum
    assert trajectory2.md_uom == 'm'


def test_Trajectory_add_well_feature_and_interp(example_model_and_crs):

    # Prepare an example Trajectory without a well feature
    wellname = "Hullabaloo"
    model, crs = example_model_and_crs
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -100),
                                md_reference = 'kelly bushing')
    datum.create_xml()
    traj = resqpy.well.Trajectory(parent_model = model, md_datum = datum, well_name = wellname)

    # Add the well interp
    assert traj.wellbore_feature is None
    assert traj.wellbore_interpretation is None
    traj.create_feature_and_interpretation()

    # Check well is present
    assert traj.wellbore_feature is not None
    assert traj.wellbore_feature.feature_name == wellname


def test_compute_from_deviation_survey(example_model_with_well):

    # --------- Arrange ----------
    # Create a Deviation Survey object in memory
    # Load example model from a fixture
    model, well_interp, datum, _ = example_model_with_well

    # Create a deviation survey
    data = dict(
        title = 'Majestic Umlaut ö',
        originator = 'Thor, god of sparkles',
        md_uom = 'ft',
        angle_uom = 'rad',
        is_final = True,
    )
    array_data = dict(
        measured_depths = np.array([1, 2, 3], dtype = float) + 1000.0,
        azimuths = np.array([4, 5, 6], dtype = float),
        inclinations = np.array([1, 2, 3], dtype = float),
        first_station = np.array([0, -1, 999], dtype = float),
    )

    survey = resqpy.well.DeviationSurvey(
        parent_model = model,
        represented_interp = well_interp,
        md_datum = datum,
        **data,
        **array_data,
    )

    # --------- Act ----------
    # Create a trajectory from the deviation survey
    trajectory_from_deviation_survey = resqpy.well.Trajectory(parent_model = model, deviation_survey = survey)

    # --------- Assert ----------
    assert trajectory_from_deviation_survey is not None
    assert trajectory_from_deviation_survey.knot_count == survey.station_count == \
           trajectory_from_deviation_survey.control_points.shape[0] == 3
    assert trajectory_from_deviation_survey.md_datum == datum
    assert trajectory_from_deviation_survey.measured_depths[0] == 1001.0


def test_load_from_wellspec(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (4, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    wellspec_file = os.path.join(model.epc_directory, 'wellspec.dat')
    well_name = 'DOGLEG'
    source_df = pd.DataFrame([[2, 2, 1, 0.0, 0.0, 0.0, 0.25], [2, 2, 2, 0.45, -90.0, 2.5, 0.25],
                              [2, 3, 2, 0.45, -90.0, 1.0, 0.20], [2, 3, 3, 0.0, 0.0, -0.5, 0.20]],
                             columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW'])
    with open(wellspec_file, 'w') as fp:
        fp.write(F'WELLSPEC {well_name}\n')
        for col in source_df.columns:
            fp.write(f' {col:>6s}')
        fp.write('\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                if col in ['IW', 'JW', 'L']:
                    fp.write(f' {int(row[col]):6d}')
                else:
                    fp.write(f' {row[col]:6.2f}')
            fp.write('\n')

    # --------- Act ----------
    # Create a trajectory from the WELLSPEC file
    trajectory_from_wellspec = resqpy.well.Trajectory(parent_model = model,
                                                      well_name = well_name,
                                                      length_uom = 'm',
                                                      grid = grid,
                                                      wellspec_file = wellspec_file)

    # --------- Assert ----------
    assert trajectory_from_wellspec is not None
    assert trajectory_from_wellspec.knot_count == len(trajectory_from_wellspec.control_points)


def test_load_from_ascii_file(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    well_names = [None, 'SnowCone']
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([300, 310, 330, 340])
    zs = mds - elevation
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [1, 2, 3, 4],
        'Y': [1, 2, 3, 4],
        'Z': zs,
        'WELL': ['SnowCone', 'SnowCone', 'Paime', 'Paime']
    })
    trajectory_data_file_path = os.path.join(model.epc_directory, 'trajectory_data.csv')
    source_dataframe.to_csv(trajectory_data_file_path)

    for well_name in well_names:
        if well_name is None:
            # --------- Act ----------
            # Create a trajectory from the csv file
            with pytest.raises(Exception) as excinfo:
                trajectory_from_ascii = resqpy.well.Trajectory(parent_model = model,
                                                               ascii_trajectory_file = trajectory_data_file_path,
                                                               length_uom = 'm',
                                                               md_datum = datum)

            # -------- Assert ---------
            assert "attempt to set trajectory for unidentified well from ascii file holding data for multiple wells" in str(
                excinfo.value)
        else:
            # --------- Act ----------
            # Create a trajectory from the csv file
            trajectory_from_ascii = resqpy.well.Trajectory(parent_model = model,
                                                           ascii_trajectory_file = trajectory_data_file_path,
                                                           well_name = well_name,
                                                           length_uom = 'm',
                                                           md_datum = datum,
                                                           set_tangent_vectors = True)
            # -------- Assert ---------
            assert trajectory_from_ascii is not None
            np.testing.assert_almost_equal(trajectory_from_ascii.tangent_vectors[0],
                                           trajectory_from_ascii.tangent_vectors[1])
            vec.isclose(trajectory_from_ascii.tangent_vectors[:, 2], np.array([1, 1]), 0.01)


def test_set_tangents(example_model_and_crs):
    # --------- Arrange ----------
    model, crs = example_model_and_crs
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([300, 310, 330, 340])
    zs = mds - elevation
    well_name = 'JubJub'
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [1, 2, 3, 4],
        'Y': [1, 2, 3, 4],
        'Z': zs,
        'WELL': ['JubJub', 'JubJub', 'JubJub', 'JubJub']
    })

    # Create a trajectory from dataframe
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')

    # --------- Act ----------
    # Calculate tangent vectors based on control points
    trajectory.set_tangents()
    # --------- Assert ----------
    assert trajectory.tangent_vectors is not None
    assert trajectory.tangent_vectors.shape == (trajectory.knot_count, 3)
    assert vec.isclose(trajectory.tangent_vectors[0], trajectory.tangent_vectors[1], tolerance = 0.01)
    np.testing.assert_equal(vec.azimuths(trajectory.tangent_vectors),
                            np.array([45, 45, 45, 45]))  # X and Y coordinates follow straight line x = y
    for v in trajectory.tangent_vectors:
        assert vec.inclination(v) < 10


def test_xyz_for_md(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([300, 310, 330, 340])
    zs = mds - elevation
    well_name = 'CoconutDrop'
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [1, 2, 3, 4],
        'Y': [1, 2, 3, 4],
        'Z': zs,
        'WELL': ['CoconutDrop', 'CoconutDrop', 'CoconutDrop', 'CoconutDrop']
    })

    # Create a trajectory from dataframe
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')
    # --------- Act ----------
    # # Get the xyz triplet for a given measured depth
    x, y, z = trajectory.xyz_for_md(md = 305)

    # # -------- Assert ---------
    assert (math.isclose(x, 1.5)) & (math.isclose(y, 1.5)) & (math.isclose(z, 205))


def test_splined_trajectory(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([300, 310, 320, 340])
    zs = mds - elevation
    well_name = 'CoconutDrop'
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [1, 2, 3, 4],
        'Y': [1, 2, 3, 4],
        'Z': zs,
        'WELL': ['CoconutDrop', 'CoconutDrop', 'CoconutDrop', 'CoconutDrop']
    })

    # Create a trajectory from dataframe
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')
    # --------- Act ----------
    # Get splined trajectory
    splined_trajectory = trajectory.splined_trajectory(well_name = 'CoconutDrop')

    # -------- Assert ---------
    assert splined_trajectory is not None
    np.testing.assert_almost_equal(trajectory.measured_depths[0], splined_trajectory.measured_depths[0], 0)
    np.testing.assert_almost_equal(trajectory.measured_depths[-1], splined_trajectory.measured_depths[-1], 0)


def test_dataframe(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([300, 310, 330, 340])
    zs = mds - elevation
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [1, 2, 3, 4],
        'Y': [1, 2, 3, 4],
        'Z': zs,
    })

    # Create a trajectory from dataframe
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        md_datum = datum,
                                        length_uom = 'm')

    # --------- Act ----------
    # Get the dataframe containing the trajectory data
    returned_dataframe = trajectory.dataframe()

    # -------- Assert ---------
    pd.testing.assert_frame_equal(source_dataframe.astype(float), returned_dataframe)



import os
import numpy as np
import pandas as pd
from numpy.testing import assert_array_almost_equal
import pytest

import resqpy.olio.uuid as bu
from resqpy.grid import RegularGrid
from resqpy.model import Model
import resqpy.property as rqp
from resqpy.well.well_utils import _derive_from_wellspec_check_grid_name
import resqpy.well


def test_wellspec_properties(example_model_and_crs):
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (3, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    wellspec_file = os.path.join(model.epc_directory, 'wellspec.dat')
    well_name = 'DOGLEG'
    source_df = pd.DataFrame([[2, 2, 1, 0.0, 0.0, 0.0, 0.25, 0.9], [2, 2, 2, 0.45, -90.0, 2.5, 0.25, 0.9],
                              [2, 3, 2, 0.45, -90.0, 1.0, 0.20, 0.9], [2, 3, 3, 0.0, 0.0, -0.5, 0.20, 0.9]],
                             columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW', 'PPERF'])
    with open(wellspec_file, 'w') as fp:
        fp.write(F'WELLSPEC {well_name}\n')
        for col in source_df.columns:
            fp.write(f' {col:>6s}')
        fp.write('\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                if col in ['IW', 'JW', 'L']:
                    fp.write(f' {int(row[col]):6d}')
                else:
                    fp.write(f' {row[col]:6.2f}')
            fp.write('\n')
    bw = resqpy.well.BlockedWell(
        model,
        wellspec_file = wellspec_file,
        well_name = well_name,
        use_face_centres = True,
        add_wellspec_properties = True,
    )
    assert bw is not None
    bw_uuid = bw.uuid
    skin_uuid = model.uuid(title = 'SKIN', related_uuid = bw.uuid)
    assert skin_uuid is not None
    skin_prop = rqp.Property(model, uuid = skin_uuid)
    assert skin_prop is not None
    assert_array_almost_equal(skin_prop.array_ref(), [0.0, 2.5, 1.0, -0.5])
    model.store_epc()
    print(model.grid().property_collection.titles())
    # re-open model from persistent storage
    model = Model(model.epc_file)
    bw2_uuid = model.uuid(obj_type = 'BlockedWellboreRepresentation', title = 'DOGLEG')
    assert bw2_uuid is not None
    bw2 = resqpy.well.BlockedWell(model, uuid = bw2_uuid)
    assert bu.matching_uuids(bw_uuid, bw2_uuid)
    df2 = bw.dataframe(extra_columns_list = ['ANGLV', 'ANGLA', 'LENGTH', 'SKIN', 'RADW', 'PPERF'],
                       use_properties = True)
    assert df2 is not None
    assert len(df2.columns) == 9
    for col in ['ANGLV', 'ANGLA', 'SKIN', 'RADW', 'PPERF']:
        assert_array_almost_equal(np.array(source_df[col]), np.array(df2[col]))
    df3 = bw.dataframe(extra_columns_list = ['ANGLV', 'ANGLA', 'LENGTH', 'SKIN', 'RADW', 'PPERF'],
                       use_properties = ['SKIN', 'RADW'],
                       perforation_list = [(125, 175)])
    for col in ['SKIN', 'RADW']:
        assert_array_almost_equal(np.array(source_df[col]), np.array(df3[col]))


@pytest.mark.parametrize('check_grid_name,name_for_check,col_list', [(True, 'BATTLESHIP', ['IW', 'JW', 'L', 'GRID']),
                                                                     (False, None, ['IW', 'JW', 'L'])])
def test_derive_from_wellspec_check_grid_name(example_model_and_crs, check_grid_name, name_for_check, col_list):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (5, 3, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       title = 'Battleship',
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    well_name = 'DOGLEG'
    col_list_orig = ['IW', 'JW', 'L']

    # --------- Act ----------
    result = _derive_from_wellspec_check_grid_name(check_grid_name = check_grid_name,
                                                   grid = grid,
                                                   col_list = col_list_orig)

    # --------- Assert ----------
    assert result[0] == name_for_check
    assert result[1] == col_list


def test_set_for_column(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (5, 3, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    well_name = 'DOGLEG'
    bw = resqpy.well.BlockedWell(model, well_name = well_name, use_face_centres = True, add_wellspec_properties = True)

    # --------- Act ----------
    # populate empty blocked well object for a 'vertical' well in the given column
    bw.set_for_column(well_name = well_name, grid = grid, col_ji0 = (1, 1))

    # --------- Assert ----------
    assert bw.cell_count == 5
    assert bw.node_count == len(
        bw.node_mds) == len(bw.trajectory.measured_depths) - 1  # tail added to trajectory measured depths
    assert (len(bw.grid_list) == 1) & (bw.grid_list[0] == grid)


def test_derive_from_cell_list(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (5, 3, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    well_name = 'DOGLEG'
    cell_kji0_list = np.array([(1, 1, 1), (2, 2, 2), (3, 3, 3), (3, 3, 4)])
    bw = resqpy.well.BlockedWell(model, well_name = well_name, use_face_centres = True, add_wellspec_properties = True)

    # --------- Act ----------
    # populate empty blocked well object for a 'vertical' well in the given column
    bw.derive_from_cell_list(cell_kji0_list = cell_kji0_list, well_name = well_name, grid = grid)

    # --------- Assert ----------
    # no tail added to trajectory measured depths as well terminates in 4th layer
    assert bw.node_count == len(bw.node_mds) == len(bw.trajectory.measured_depths)
    assert bw.cell_count == len(cell_kji0_list)


def test_grid_uuid_list(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (5, 3, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    grid_uuid = grid.uuid
    well_name = 'DOGLEG'
    cell_kji0_list = np.array([(1, 1, 1), (2, 2, 2), (3, 3, 3), (3, 3, 4)])
    bw = resqpy.well.BlockedWell(model, well_name = well_name, use_face_centres = True, add_wellspec_properties = True)

    # --------- Act ----------
    # populate empty blocked well object for a 'vertical' well in the given column
    bw.derive_from_cell_list(cell_kji0_list = cell_kji0_list, well_name = well_name, grid = grid)
    # get list of grid uuids associated with this blocked well
    grid_uuid_list = bw.grid_uuid_list()

    # --------- Assert ----------
    assert len(grid_uuid_list) == 1
    assert bu.matching_uuids(grid_uuid_list[0], grid_uuid)


def test_verify_grid_name(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (3, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    well_name = 'DOGLEG'
    source_df = pd.DataFrame(
        [[2, 2, 1, 0.0, 0.0, 0.0, 0.25, 0.9, 'grid_1'], [2, 2, 2, 0.45, -90.0, 2.5, 0.25, 0.9, 'grid_1'],
         [2, 3, 2, 0.45, -90.0, 1.0, 0.20, 0.9, 'grid_2'], [2, 3, 3, 0.0, 0.0, -0.5, 0.20, 0.9, 'grid_1']],
        columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW', 'PPERF', 'GRID'])
    row = source_df.iloc[2]
    bw = resqpy.well.BlockedWell(model, well_name = well_name, use_face_centres = True)

    # --------- Act ----------
    skipped_warning_grid, skip_row = bw._BlockedWell__verify_grid_name(grid_name_to_check = 'grid_1',
                                                                       row = row,
                                                                       skipped_warning_grid = None,
                                                                       well_name = well_name)

    # --------- Assert ----------
    assert skipped_warning_grid == 'grid_2'
    assert skip_row


def test_calculate_exit_and_entry(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (3, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    cp = grid.corner_points(cell_kji0 = (2, 2, 1))
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    well_name = 'DOGLEG'
    source_df = pd.DataFrame(
        [[2, 2, 1, 0.0, 0.0, 0.0, 0.25, 0.9, 'grid_1'], [2, 2, 2, 0.45, -90.0, 2.5, 0.25, 0.9, 'grid_1'],
         [2, 3, 2, 0.45, -90.0, 1.0, 0.20, 0.9, 'grid_1'], [2, 3, 3, 0.0, 0.0, -0.5, 0.20, 0.9, 'grid_1']],
        columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW', 'PPERF', 'GRID'])
    row = source_df.iloc[0]
    bw = resqpy.well.BlockedWell(model, well_name = well_name, use_face_centres = True)

    # --------- Act ----------
    (entry_axis, entry_polarity, entry_xyz, exit_axis, exit_polarity, exit_xyz) = bw._BlockedWell__calculate_entry_and_exit_axes_polarities_and_points_using_angles\
        (row = row, cp = cp, well_name = well_name)
    print((entry_axis, entry_polarity, entry_xyz, exit_axis, exit_polarity, exit_xyz))

    # --------- Assert ----------
    # TODO: less trivial assertion
    for value in (entry_axis, entry_polarity, entry_xyz, exit_axis, exit_polarity, exit_xyz):
        assert value is not None


def test_calculate_cell_cp_center_and_vectors(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (3, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    cp_expected = grid.corner_points(cell_kji0 = (2, 2, 1))
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    well_name = 'DOGLEG'
    source_df = pd.DataFrame(
        [[2, 2, 1, 0.0, 0.0, 0.0, 0.25, 0.9, 'grid_1'], [2, 2, 2, 0.45, -90.0, 2.5, 0.25, 0.9, 'grid_1'],
         [2, 3, 2, 0.45, -90.0, 1.0, 0.20, 0.9, 'grid_1'], [2, 3, 3, 0.0, 0.0, -0.5, 0.20, 0.9, 'grid_1']],
        columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW', 'PPERF', 'GRID'])
    bw = resqpy.well.BlockedWell(model, well_name = well_name, use_face_centres = True)

    # --------- Act ----------
    cp, cell_centre, entry_vector, exit_vector = bw._BlockedWell__calculate_cell_cp_center_and_vectors(
        grid = grid,
        cell_kji0 = (2, 2, 1),
        entry_xyz = np.array([75, -125, 200]),
        exit_xyz = np.array([100, -100, 250]),
        well_name = well_name)

    # --------- Assert ----------
    np.testing.assert_equal(cp, cp_expected)
    np.testing.assert_equal(cell_centre, np.array([75., -125., 225.]))
    np.testing.assert_equal(entry_vector, np.array([0., 0., -2500.]))
    np.testing.assert_equal(exit_vector, np.array([2500., 2500., 2500.]))


def test_import_from_cellio_file(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (3, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)

    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    grid_uuid = grid.uuid
    cellio_file = os.path.join(model.epc_directory, 'cellio.dat')
    well_name = 'Banoffee'
    source_df = pd.DataFrame(
        [[2, 2, 1, 25, -25, 125, 26, -26, 126], [2, 2, 2, 26, -26, 126, 27, -27, 127],
         [2, 2, 3, 27, -27, 127, 28, -28, 128]],
        columns = ['i_index', 'j_index', 'k_index', 'x_in', 'y_in', 'z_in', 'x_out', 'y_out', 'z_out'])

    with open(cellio_file, 'w') as fp:
        fp.write('1.0\n')
        fp.write('Undefined\n')
        fp.write(f'{well_name}\n')
        fp.write('9\n')
        for col in source_df.columns:
            fp.write(f' {col}\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                fp.write(f' {int(row[col])}')
            fp.write('\n')


# --------- Act ----------
    bw = resqpy.well.BlockedWell(model, use_face_centres = True)
    # assert that certain attributes have not been populated
    assert bw.grid_list == []
    assert bw.trajectory is None
    assert bw.cell_count is None
    assert bw.node_count is None
    assert bw.node_mds is None

    bw.import_from_rms_cellio(cellio_file = cellio_file, well_name = well_name, grid = grid)

    # --------- Assert ----------
    assert bw.grid_list[0].uuid == grid_uuid
    assert bw.trajectory is not None
    assert bw.cell_count == 3  # 3 lines in the cellio file
    assert bw.node_count == len(bw.node_mds) == 4  # added tail to trajectory


@pytest.mark.parametrize('ntg_multiplier,length_mode,status', [(0.5, 'straight', 'OFF'), (1, 'MD', 'ON')])
def test_dataframe(example_model_and_crs, ntg_multiplier, length_mode, status):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (5, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    perm_i_array = np.random.random(grid.extent_kji)
    ntg_array = np.ones(grid.extent_kji) * ntg_multiplier
    perm_i_prop = rqp.Property.from_array(model,
                                          perm_i_array,
                                          source_info = 'random',
                                          keyword = 'PERMI',
                                          support_uuid = grid.uuid,
                                          property_kind = 'permeability rock',
                                          indexable_element = 'cells',
                                          uom = 'Euc')
    ntg_prop = rqp.Property.from_array(model,
                                       ntg_array,
                                       keyword = 'NTG',
                                       source_info = 'random',
                                       support_uuid = grid.uuid,
                                       property_kind = 'net to gross ratio',
                                       indexable_element = 'cells',
                                       uom = 'Euc')
    perm_i_prop.write_hdf5()
    perm_i_prop.create_xml()
    ntg_prop.write_hdf5()
    ntg_prop.create_xml()
    perm_i_uuid = perm_i_prop.uuid
    ntg_uuid = ntg_prop.uuid
    wellspec_file = os.path.join(model.epc_directory, 'wellspec.dat')
    well_name = 'DOGLEG'
    source_df = pd.DataFrame(
        [[2, 2, 1, 0.0, 0.0, 0.0, 0.25], [2, 2, 2, 45, -90.0, 2.5, 0.25], [2, 3, 2, 45, -90.0, 1.0, 0.20],
         [2, 3, 3, 45, -90.0, -0.5, 0.20], [2, 3, 4, 45, -90.0, 1.1, 0.20], [2, 3, 5, 0.0, 0.0, 1.0, 0.20]],
        columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW'])
    with open(wellspec_file, 'w') as fp:
        fp.write(F'WELLSPEC {well_name}\n')
        for col in source_df.columns:
            fp.write(f' {col:>6s}')
        fp.write('\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                if col in ['IW', 'JW', 'L']:
                    fp.write(f' {int(row[col]):6d}')
                else:
                    fp.write(f' {row[col]:6.2f}')
            fp.write('\n')
    bw = resqpy.well.BlockedWell(model,
                                 wellspec_file = wellspec_file,
                                 well_name = well_name,
                                 use_face_centres = True,
                                 add_wellspec_properties = True)

    # --------- Act ----------
    df = bw.dataframe(extra_columns_list = ['X', 'Y', 'DEPTH', 'ANGLV', 'ANGLA', 'SKIN', 'RADW', 'KH', 'WI', 'WBC'],
                      add_as_properties = True,
                      perforation_list = [(125, 320)],
                      max_depth = 500,
                      perm_i_uuid = perm_i_uuid,
                      perm_j_uuid = perm_i_uuid,
                      ntg_uuid = ntg_uuid,
                      preferential_perforation = True,
                      stat = status,
                      min_k0 = 1,
                      max_k0 = 5,
                      use_face_centres = True,
                      length_uom = 'm',
                      length_mode = length_mode)

    # --------- Assert ----------
    # print(df)
    assert all(df['KH'])  # successfully added a KH column as an i-direction permeability array was specified
    assert set(df['STAT']) == {status}
    assert all(df['WI'])
    assert all(df['WBC'])


def test_dataframe_from_trajectory(example_model_and_crs):
    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (5, 4, 4),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)

    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    elevation = 100
    # Create a measured depth datum
    datum = resqpy.well.MdDatum(parent_model = model,
                                crs_uuid = crs.uuid,
                                location = (0, 0, -elevation),
                                md_reference = 'kelly bushing')
    mds = np.array([100, 210, 230, 240, 250])
    zs = mds - elevation
    well_name = 'CoconutDrop'
    source_dataframe = pd.DataFrame({
        'MD': mds,
        'X': [25, 50, 75, 100, 100],
        'Y': [25, -50, -75, -100, -100],
        'Z': zs,
        'WELL': ['CoconutDrop', 'CoconutDrop', 'CoconutDrop', 'CoconutDrop', 'CoconutDrop']
    })

    # Create a trajectory from dataframe
    trajectory = resqpy.well.Trajectory(parent_model = model,
                                        data_frame = source_dataframe,
                                        well_name = well_name,
                                        md_datum = datum,
                                        length_uom = 'm')
    trajectory.write_hdf5()
    trajectory.create_xml()
    bw = resqpy.well.BlockedWell(model,
                                 well_name = well_name,
                                 grid = grid,
                                 trajectory = trajectory,
                                 use_face_centres = True)

    # --------- Act ----------
    df = bw.dataframe(extra_columns_list = ['X', 'Y', 'DEPTH', 'RADW'], stat = 'ON', length_uom = 'ft')

    # --------- Assert ----------
    assert bw.trajectory is not None
    assert set(df['RADW']) == {0.25}
    assert set(df['STAT']) == {'ON'}
    assert all(df['X']) == all(df['Y']) == all(df['DEPTH'])  # successfully got xyz points from trajectory and
    # converted them to ft


def test_write_wellspec(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (3, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    perm_array = np.random.random(grid.extent_kji)
    perm_prop = rqp.Property.from_array(model,
                                        perm_array,
                                        source_info = 'random',
                                        keyword = 'PERMI',
                                        support_uuid = grid.uuid,
                                        property_kind = 'permeability',
                                        indexable_element = 'cells',
                                        uom = 'Euc')
    perm_prop.write_hdf5()
    perm_prop.create_xml()
    perm_uuid = perm_prop.uuid
    wellspec_file = os.path.join(model.epc_directory, 'wellspec.dat')
    well_name = 'DOGLEG'
    source_df = pd.DataFrame([[2, 2, 1, 0.0, 0.0, 0.0, 0.25], [2, 2, 2, 45, -90.0, 2.5, 0.25],
                              [2, 3, 2, 45, -90.0, 1.0, 0.20], [2, 3, 3, 0.0, 0.0, -0.5, 0.20]],
                             columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW'])
    with open(wellspec_file, 'w') as fp:
        fp.write(F'WELLSPEC {well_name}\n')
        for col in source_df.columns:
            fp.write(f' {col:>6s}')
        fp.write('\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                if col in ['IW', 'JW', 'L']:
                    fp.write(f' {int(row[col]):6d}')
                else:
                    fp.write(f' {row[col]:6.2f}')
            fp.write('\n')
    bw = resqpy.well.BlockedWell(model,
                                 wellspec_file = wellspec_file,
                                 well_name = well_name,
                                 use_face_centres = True,
                                 add_wellspec_properties = True)

    # --------- Act ----------
    df = bw.dataframe(extra_columns_list = ['ANGLV', 'ANGLA', 'SKIN', 'RADW', 'KH'],
                      add_as_properties = True,
                      perm_i_uuid = perm_uuid)

    wellspec_file2 = os.path.join(model.epc_directory, 'wellspec2.dat')
    df2 = bw.write_wellspec(wellspec_file = wellspec_file2,
                            well_name = well_name,
                            extra_columns_list = ['ANGLV', 'ANGLA', 'SKIN', 'RADW'],
                            length_uom = 'm',
                            length_uom_comment = '?')

    # --------- Assert ----------
    pd.testing.assert_frame_equal(df[['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW']], df2, check_dtype = False)
    # TODO find out why initially when ANGLV was 0.45, the Blocked Well dataframe method changed the values to 45
    # TODO find out why AngleA values of 0 transformed to nan values?


def test_convenience_methods_xyz_and_kji0_marker(example_model_and_crs):

    # --------- Arrange ----------
    model, crs = example_model_and_crs
    grid = RegularGrid(model,
                       extent_kji = (3, 4, 3),
                       dxyz = (50.0, -50.0, 50.0),
                       origin = (0.0, 0.0, 100.0),
                       crs_uuid = crs.uuid,
                       set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    wellspec_file = os.path.join(model.epc_directory, 'wellspec.dat')
    well_name = 'DOGLEG'
    source_df = pd.DataFrame([[1, 2, 2, 0.0, 0.0, 0.0, 0.25], [2, 2, 2, 45, -90.0, 2.5, 0.25],
                              [2, 3, 2, 45, -90.0, 1.0, 0.20], [2, 3, 3, 0.0, 0.0, -0.5, 0.20]],
                             columns = ['IW', 'JW', 'L', 'ANGLV', 'ANGLA', 'SKIN', 'RADW'])
    with open(wellspec_file, 'w') as fp:
        fp.write(F'WELLSPEC {well_name}\n')
        for col in source_df.columns:
            fp.write(f' {col:>6s}')
        fp.write('\n')
        for row_index in range(len(source_df)):
            row = source_df.iloc[row_index]
            for col in source_df.columns:
                if col in ['IW', 'JW', 'L']:
                    fp.write(f' {int(row[col]):6d}')
                else:
                    fp.write(f' {row[col]:6.2f}')
            fp.write('\n')
    bw = resqpy.well.BlockedWell(model,
                                 wellspec_file = wellspec_file,
                                 well_name = well_name,
                                 use_face_centres = True,
                                 add_wellspec_properties = True)

    # --------- Act ----------
    cell_0, grid_0_uuid = bw.kji0_marker()  # cell_0 returned in the format (k0, j0, i0)
    _, crs_uuid = bw.xyz_marker()

    # --------- Assert ----------
    np.testing.assert_equal(cell_0, np.array([1, 1, 0]))  # start blocked well at (iw, jw, l) == (1, 2, 2)
    assert grid_0_uuid == grid.uuid
    assert crs_uuid == crs.uuid  # TODO: less trivial assertion?



"""High level unsplit grid function."""

import logging

log = logging.getLogger(__name__)

import os

import resqpy.model as rq
import resqpy.olio.xml_et as rqet

from resqpy.derived_model._common import _prepare_simple_inheritance, _write_grid, _establish_model_and_source_grid
from resqpy.derived_model._copy_grid import copy_grid


def unsplit_grid(epc_file,
                 source_grid = None,
                 inherit_properties = False,
                 inherit_realization = None,
                 inherit_all_realizations = False,
                 new_grid_title = None,
                 new_epc_file = None):
    """Extends epc file with a new grid which is a version of the source grid with all faults healed.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the unsplit grid (& crs)

    returns:
       a new grid (grid.Grid object) which is an unfaulted copy of the source grid

    notes:
       the faults are healed by shifting the thrown sides up and down to the midpoint, only along the line of the fault;
       to smooth the adjustments away from the line of the fault, use the global_fault_throw_scaling() function first
    """

    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    model, source_grid = _establish_model_and_source_grid(epc_file, source_grid)
    assert source_grid.grid_representation == 'IjkGrid'
    assert model is not None

    assert source_grid.has_split_coordinate_lines, 'source grid is unfaulted'

    # take a copy of the grid
    grid = copy_grid(source_grid, model)

    if grid.inactive is not None:
        log.debug('copied grid inactive shape: ' + str(grid.inactive.shape))

    # heal faults in the grid
    grid.cache_all_geometry_arrays()  # probably already cached anyway
    unsplit = source_grid.unsplit_points_ref()
    grid.points_cached = unsplit.copy()
    assert grid.points_cached.shape == (grid.nk + 1, grid.nj + 1, grid.ni + 1, 3), 'unsplit points have incorrect shape'

    grid.has_split_coordinate_lines = False
    delattr(grid, 'split_pillar_indices_cached')
    delattr(grid, 'cols_for_split_pillars')
    delattr(grid, 'cols_for_split_pillars_cl')
    if hasattr(grid, 'pillars_for_column'):
        delattr(grid, 'pillars_for_column')

    collection = _prepare_simple_inheritance(grid, source_grid, inherit_properties, inherit_realization,
                                             inherit_all_realizations)
    # todo: recompute depth properties (and volumes, cell lengths etc. if being strict)

    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'unfaulted version of ' + str(rqet.citation_title_for_node(source_grid.root))

    # write model
    if new_epc_file:
        _write_grid(new_epc_file, grid, property_collection = collection, grid_title = new_grid_title, mode = 'w')
    else:
        ext_uuid, _ = model.h5_uuid_and_path_for_node(rqet.find_nested_tags(source_grid.root, ['Geometry', 'Points']),
                                                      'Coordinates')
        _write_grid(epc_file,
                    grid,
                    ext_uuid = ext_uuid,
                    property_collection = collection,
                    grid_title = new_grid_title,
                    mode = 'a')

    return grid



"""DEPRECATED, use resqpy.weights_and_measures instead."""

import warnings

from resqpy.weights_and_measures import *  # noqa

warnings.warn(
    "resqpy.olio.weights_and_measures is Deprecated and will be removed. Please update your code to use resqpy.weights_and_measures"
)



"""_stratigraphic_unit_interpretation.py: RESQML StratigraphicUnitInterpretation class."""

version = '24th November 2021'

# NB: in this module, the term 'unit' refers to a geological stratigraphic unit, i.e. a layer of rock, not a unit of measure

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.weights_and_measures as wam
from resqpy.olio.xml_namespaces import curly_namespace as ns
from resqpy.strata._strata_common import valid_deposition_modes
from resqpy.strata._geologic_unit_interpretation import GeologicUnitInterpretation
from resqpy.strata._stratigraphic_unit_feature import StratigraphicUnitFeature


class StratigraphicUnitInterpretation(GeologicUnitInterpretation):
    """Class for RESQML Stratigraphic Unit Interpretation objects.

    RESQML documentation:

       Interpretation of a stratigraphic unit which includes the knowledge of the top, the bottom,
       the deposition mode.
    """

    resqml_type = 'StratigraphicUnitInterpretation'

    def __init__(
            self,
            parent_model,
            uuid = None,
            title = None,
            domain = 'time',  # or should this be depth?
            stratigraphic_unit_feature = None,
            composition = None,
            material_implacement = None,
            deposition_mode = None,
            min_thickness = None,
            max_thickness = None,
            thickness_uom = None,
            extra_metadata = None):
        """Initialises a stratigraphic unit interpretation object.

        arguments:
           parent_model (model.Model): the model with which the new interpretation will be associated
           uuid (uuid.UUID, optional): the uuid of an existing RESQML stratigraphic unit interpretation from which
              this object will be initialised
           title (str, optional): the citation title (feature name) of the new interpretation;
              ignored if uuid is not None
           domain (str, default 'time'): 'time', 'depth' or 'mixed', being the domain of the interpretation;
              ignored if uuid is not None
           stratigraphic_unit_feature (StratigraphicUnitFeature, optional): the feature which this object is
              an interpretation of; ignored if uuid is not None
           composition (str, optional): the interpreted composition of the stratigraphic unit; if present, must be
              in valid_compositions; ignored if uuid is not None
           material_implacement (str, optional): the interpeted material implacement of the stratigraphic unit;
              if present, must be in valid_implacements, ie. 'autochtonous' or 'allochtonous';
              ignored if uuid is not None
           deposition_mode (str, optional): indicates whether deposition within the unit is interpreted as parallel
              to top, base or another boundary, or is proportional to thickness; if present, must be in
              valid_deposition_modes; ignored if uuid is not None
           min_thickness (float, optional): the minimum thickness of the unit; ignored if uuid is not None
           max_thickness (float, optional): the maximum thickness of the unit; ignored if uuid is not None
           thickness_uom (str, optional): the length unit of measure of the minimum and maximum thickness; required
              if either thickness argument is provided and uuid is None; if present, must be a valid length uom
           extra_metadata (dict, optional): extra metadata items for the new interpretation

        returns:
           a new stratigraphic unit interpretation resqpy object

        notes:
           if given, the thickness_uom must be a valid RESQML length unit of measure; the set of valid uoms is
           returned by: weights_and_measures.valid_uoms(quantity = 'length');
           the RESQML 2.0.1 schema definition includes a spurious trailing space in the names of two compositions;
           resqpy removes such spaces in the composition attribute as presented to calling code (but includes them
           in xml)
        """

        self.deposition_mode = deposition_mode
        self.min_thickness = min_thickness
        self.max_thickness = max_thickness
        self.thickness_uom = thickness_uom
        super().__init__(parent_model,
                         uuid = uuid,
                         title = title,
                         domain = domain,
                         geologic_unit_feature = stratigraphic_unit_feature,
                         composition = composition,
                         material_implacement = material_implacement,
                         extra_metadata = extra_metadata)
        if self.deposition_mode is not None:
            assert self.deposition_mode in valid_deposition_modes
        if self.min_thickness is not None or self.max_thickness is not None:
            assert self.thickness_uom in wam.valid_uoms(quantity = 'length')

    @property
    def stratigraphic_unit_feature(self):
        """Returns the interpreted geologic unit feature."""

        return self.geologic_unit_feature

    def _load_from_xml(self):
        """Loads class specific attributes from xml for an existing RESQML object; called from BaseResqpy."""
        super()._load_from_xml()
        root_node = self.root
        assert root_node is not None
        feature_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(root_node, ['InterpretedFeature', 'UUID']))
        if feature_uuid is not None:
            self.geologic_unit_feature = StratigraphicUnitFeature(self.model,
                                                                  uuid = feature_uuid,
                                                                  title = self.model.title(uuid = feature_uuid))
        # load deposition mode and min & max thicknesses (& uom), if present
        self.deposition_mode = rqet.find_tag_text(root_node, 'DepositionMode')
        for min_max in ['Min', 'Max']:
            thick_node = rqet.find_tag(root_node, min_max + 'Thickness')
            if thick_node is not None:
                thick = float(thick_node.text)
                if min_max == 'Min':
                    self.min_thickness = thick
                else:
                    self.max_thickness = thick
                thick_uom = thick_node.attrib['uom']  # todo: check this is correct uom representation
                if self.thickness_uom is None:
                    self.thickness_uom = thick_uom
                else:
                    assert thick_uom == self.thickness_uom, 'inconsistent length units of measure for stratigraphic thicknesses'

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False.

        arguments:
           other (StratigraphicUnitInterpretation): the other interpretation to compare this one against
           check_extra_metadata (bool, default True): if True, then extra metadata items must match for the two
              interpretations to be deemed equivalent; if False, extra metadata is ignored in the comparison

        returns:
           bool: True if this interpretation is essentially the same as the other; False otherwise
        """
        if not super().is_equivalent(other):
            return False
        if self.deposition_mode is not None and other.deposition_mode is not None:
            return self.deposition_mode == other.deposition_mode
        # note: thickness range information might be lost as not deemed of significance in comparison
        return True

    def create_xml(self, add_as_part = True, add_relationships = True, originator = None, reuse = True):
        """Creates a stratigraphic unit interpretation xml tree.

        arguments:
           add_as_part (bool, default True): if True, the interpretation is added to the parent model as a high level part
           add_relationships (bool, default True): if True and add_as_part is True, a relationship is created with
              the referenced stratigraphic unit feature
           originator (str, optional): if present, is used as the originator field of the citation block
           reuse (bool, default True): if True, the parent model is inspected for any equivalent interpretation and, if found,
              the uuid of this interpretation is set to that of the equivalent part

        returns:
           lxml.etree._Element: the root node of the newly created xml tree for the interpretation
        """

        if reuse and self.try_reuse():
            return self.root

        sui = super().create_xml(add_as_part = add_as_part,
                                 add_relationships = add_relationships,
                                 originator = originator,
                                 reuse = False)
        assert sui is not None

        if self.deposition_mode is not None:
            assert self.deposition_mode in valid_deposition_modes,  \
               f'invalid deposition mode {self.deposition_mode} for stratigraphic unit interpretation'
            dm_node = rqet.SubElement(sui, ns['resqml2'] + 'DepositionMode')
            dm_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'DepositionMode')
            dm_node.text = self.deposition_mode

        if self.min_thickness is not None or self.max_thickness is not None:
            assert self.thickness_uom in wam.valid_uoms(quantity = 'length')

        if self.min_thickness is not None:
            min_thick_node = rqet.SubElement(sui, ns['resqml2'] + 'MinThickness')
            min_thick_node.set(ns['xsi'] + 'type', ns['eml'] + 'LengthMeasure')
            min_thick_node.set('uom', self.thickness_uom)  # todo: check this
            min_thick_node.text = str(self.min_thickness)

        if self.max_thickness is not None:
            max_thick_node = rqet.SubElement(sui, ns['resqml2'] + 'MaxThickness')
            max_thick_node.set(ns['xsi'] + 'type', ns['eml'] + 'LengthMeasure')
            max_thick_node.set('uom', self.thickness_uom)
            max_thick_node.text = str(self.max_thickness)

        return sui



"""_grid_from_cp.py: Module to generate a RESQML grid object from an input corner point array."""

version = '15th November 2021'

import logging

log = logging.getLogger(__name__)

import numpy as np
import numpy.ma as ma

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.olio.vector_utilities as vec


def grid_from_cp(model,
                 cp_array,
                 crs_uuid,
                 active_mask = None,
                 geometry_defined_everywhere = True,
                 treat_as_nan = None,
                 dot_tolerance = 1.0,
                 morse_tolerance = 5.0,
                 max_z_void = 0.1,
                 split_pillars = True,
                 split_tolerance = 0.01,
                 ijk_handedness = 'right',
                 known_to_be_straight = False):
    """Create a resqpy.grid.Grid object from a 7D corner point array.

    Arguments:
        model (resqpy.model.Model): model to which the grid will be added
        cp_array (numpy float array): 7 dimensional numpy array of nexus corner point data, in nexus ordering
        crs_uuid (uuid.UUID): uuid for the coordinate reference system
        active_mask (3d numpy bool array): array indicating which cells are active
        geometry_defined_everywhere (bool, default True): if False then inactive cells are marked as not having geometry
        treat_as_nan (float, default None): if a value is provided corner points with this value will be assigned nan
        dot_tolerance (float, default 1.0): minimum manhatten distance of primary diagonal of cell, below which cell is treated as inactive
        morse_tolerance (float, default 5.0): maximum ratio of i and j face vector lengths, beyond which cells are treated as inactive
        max_z_void (float, default 0.1): maximum z gap between vertically neighbouring corner points. Vertical gaps greater than this will introduce k gaps into resqml grid. Units are corp z units
        split_pillars (bool, default True): if False an unfaulted grid will be generated
        split_tolerance (float, default 0.01): maximum distance between neighbouring corner points before a pillar is considered 'split'. Applies to each of x, y, z differences
        ijk_handedness (str, default 'right'): 'right' or 'left'
        known_to_be_straight (bool, default False): if True pillars are forced to be straight

    notes:
       this function sets up all the geometry arrays in memory but does not write to hdf5 nor create xml: use Grid methods;
       geometry_defined_everywhere is deprecated, use treat_as_nan instead
    """
    grid = _GridFromCp(model, cp_array, crs_uuid, active_mask, geometry_defined_everywhere, treat_as_nan, dot_tolerance,
                       morse_tolerance, max_z_void, split_pillars, split_tolerance, ijk_handedness,
                       known_to_be_straight)

    return grid.grid


class _GridFromCp:
    """Class to build a resqpy grid from a Nexus CORP array"""

    def __init__(self,
                 model,
                 cp_array,
                 crs_uuid,
                 active_mask = None,
                 geometry_defined_everywhere = True,
                 treat_as_nan = None,
                 dot_tolerance = 1.0,
                 morse_tolerance = 5.0,
                 max_z_void = 0.1,
                 split_pillars = True,
                 split_tolerance = 0.01,
                 ijk_handedness = 'right',
                 known_to_be_straight = False):
        """Class to build a resqpy grid from a Nexus CORP array"""

        self.__model = model
        self.__cp_array = cp_array
        self.__crs_uuid = crs_uuid
        self.__active_mask = active_mask
        self.__geometry_defined_everywhere = geometry_defined_everywhere
        self.__treat_as_nan = treat_as_nan
        self.__dot_tolerance = dot_tolerance
        self.__morse_tolerance = morse_tolerance
        self.__max_z_void = max_z_void
        self.__split_pillars = split_pillars
        self.__split_tolerance = split_tolerance
        self.__ijk_handedness = ijk_handedness
        self.__known_to_be_straight = known_to_be_straight

        self.create_grid()

    def __get_treat_as_nan(self):
        if self.__treat_as_nan is None:
            if not self.__geometry_defined_everywhere:
                self.__treat_as_nan = 'morse'
        else:
            assert self.__treat_as_nan in ['none', 'dots', 'ij_dots', 'morse', 'inactive']
            if self.__treat_as_nan == 'none':
                self.__treat_as_nan = None

    def __get_extents(self):
        self.__nk, self.__nj, self.__ni = self.__cp_array.shape[:3]
        self.__nk_plus_1 = self.__nk + 1
        self.__nj_plus_1 = self.__nj + 1
        self.__ni_plus_1 = self.__ni + 1

    def __get_active_inactive_masks(self):
        if self.__active_mask is None:
            self.__active_mask = np.ones((self.__nk, self.__nj, self.__ni), dtype = 'bool')
            self.__inactive_mask = np.zeros((self.__nk, self.__nj, self.__ni), dtype = 'bool')
        else:
            assert self.__active_mask.shape == (self.__nk, self.__nj, self.__ni)
            self.__inactive_mask = np.logical_not(self.__active_mask)
        self.__all_active = np.all(self.__active_mask)

    def __get_dot_mask_dots(self):
        # for speed, only check primary diagonal of cells
        log.debug('geometry for cells with no length to primary cell diagonal being set to NaN')
        self.__dot_mask = np.all(
            np.abs(self.__cp_array[:, :, :, 1, 1, 1] - self.__cp_array[:, :, :, 0, 0, 0]) < self.__dot_tolerance,
            axis = -1)

    def __get_dot_mask_ijdots_or_morse(self):
        # check one diagonal of each I & J face
        log.debug(
            'geometry being set to NaN for inactive cells with no length to primary face diagonal for any I or J face')
        self.__dot_mask = np.zeros((self.__nk, self.__nj, self.__ni), dtype = bool)
        #              k_face_vecs = cp_array[:, :, :, :, 1, 1] - cp_array[:, :, :, :, 0, 0]
        j_face_vecs = self.__cp_array[:, :, :, 1, :, 1] - self.__cp_array[:, :, :, 0, :, 0]
        i_face_vecs = self.__cp_array[:, :, :, 1, 1, :] - self.__cp_array[:, :, :, 0, 0, :]
        self.__dot_mask[:] = np.where(np.all(np.abs(j_face_vecs[:, :, :, 0]) < self.__dot_tolerance, axis = -1), True,
                                      self.__dot_mask)
        self.__dot_mask[:] = np.where(np.all(np.abs(j_face_vecs[:, :, :, 1]) < self.__dot_tolerance, axis = -1), True,
                                      self.__dot_mask)
        self.__dot_mask[:] = np.where(np.all(np.abs(i_face_vecs[:, :, :, 0]) < self.__dot_tolerance, axis = -1), True,
                                      self.__dot_mask)
        self.__dot_mask[:] = np.where(np.all(np.abs(i_face_vecs[:, :, :, 1]) < self.__dot_tolerance, axis = -1), True,
                                      self.__dot_mask)
        log.debug(f'dot mask set for {np.count_nonzero(self.__dot_mask)} cells')

        if self.__treat_as_nan == 'morse':
            self.__get_dot_mask_morse(i_face_vecs, j_face_vecs)

    def __get_dot_mask_morse(self, i_face_vecs, j_face_vecs):
        morse_tol_sqr = self.__morse_tolerance * self.__morse_tolerance
        # compare face vecs lengths in xy against max for active cells: where much greater set to NaN
        len_j_face_vecs_sqr = np.sum(j_face_vecs[..., :2] * j_face_vecs[..., :2], axis = -1)
        len_i_face_vecs_sqr = np.sum(i_face_vecs[..., :2] * i_face_vecs[..., :2], axis = -1)
        dead_mask = self.__inactive_mask.reshape(self.__nk, self.__nj, self.__ni, 1).repeat(2, -1)
        #                  mean_len_active_j_face_vecs_sqr = np.mean(ma.masked_array(len_j_face_vecs_sqr, mask = dead_mask))
        #                  mean_len_active_i_face_vecs_sqr = np.mean(ma.masked_array(len_i_face_vecs_sqr, mask = dead_mask))
        max_len_active_j_face_vecs_sqr = np.max(ma.masked_array(len_j_face_vecs_sqr, mask = dead_mask))
        max_len_active_i_face_vecs_sqr = np.max(ma.masked_array(len_i_face_vecs_sqr, mask = dead_mask))
        self.__dot_mask = np.where(
            np.any(len_j_face_vecs_sqr > morse_tol_sqr * max_len_active_j_face_vecs_sqr, axis = -1), True,
            self.__dot_mask)
        self.__dot_mask = np.where(
            np.any(len_i_face_vecs_sqr > morse_tol_sqr * max_len_active_i_face_vecs_sqr, axis = -1), True,
            self.__dot_mask)
        log.debug(f'morse mask set for {np.count_nonzero(self.__dot_mask)} cells')

    def __get_nan_mask(self):
        if self.__all_active and self.__geometry_defined_everywhere:
            self.__cp_nan_mask = None
        else:
            self.__cp_nan_mask = np.any(np.isnan(self.__cp_array), axis = (3, 4, 5, 6))  # ie. if any nan per cell
            if not self.__geometry_defined_everywhere and not self.__all_active:
                if self.__treat_as_nan == 'inactive':
                    log.debug('all inactive cell geometry being set to NaN')
                    self.__cp_nan_mask = np.logical_or(self.__cp_nan_mask, self.__inactive_mask)
                else:
                    if self.__treat_as_nan == 'dots':
                        self.__get_dot_mask_dots()
                    elif self.__treat_as_nan in ['ij_dots', 'morse']:
                        self.__get_dot_mask_ijdots_or_morse()
                    else:
                        raise Exception('code broken')
                    self.__cp_nan_mask = np.logical_or(self.__cp_nan_mask,
                                                       np.logical_and(self.__inactive_mask, self.__dot_mask))
            self.__geometry_defined_everywhere = not np.any(self.__cp_nan_mask)
            if self.__geometry_defined_everywhere:
                self.__cp_nan_mask = None

    def __get_masked_cp_array(self):
        # set up masked version of corner point data based on cells with defined geometry
        if self.__geometry_defined_everywhere:
            full_mask = None
            self.__masked_cp_array = ma.masked_array(self.__cp_array, mask = ma.nomask)
            log.info('geometry present for all cells')
        else:
            full_mask = self.__cp_nan_mask.reshape((self.__nk, self.__nj, self.__ni, 1)).repeat(24, axis = 3).reshape(
                (self.__nk, self.__nj, self.__ni, 2, 2, 2, 3))
            self.__masked_cp_array = ma.masked_array(self.__cp_array, mask = full_mask)
            log.info('number of cells without geometry: ' + str(np.count_nonzero(self.__cp_nan_mask)))

    def __check_for_kgaps(self):
        self.__k_gaps = None
        self.__k_gap_raw_index = None
        self.__k_gap_after_layer = None

        if self.__nk > 1:
            # check for (vertical) voids, or un-pillar-like anomalies, which will require k gaps in the resqml ijk grid
            log.debug('checking for voids')
            gap = self.__masked_cp_array[1:, :, :, 0, :, :, :] - self.__masked_cp_array[:-1, :, :, 1, :, :, :]
            max_gap_by_layer_and_xyz = np.max(np.abs(gap), axis = (1, 2, 3, 4))
            max_gap = np.max(max_gap_by_layer_and_xyz)
            log.debug('maximum void distance: {0:.3f}'.format(max_gap))
            if max_gap > self.__max_z_void:
                self.__get_kgaps_details(max_gap_by_layer_and_xyz, gap)
            elif max_gap > 0.0:
                self.__close_gaps(gap)

    def __get_kgaps_details(self, max_gap_by_layer_and_xyz, gap):
        log.warning('maximum void distance exceeds limit, grid will include k gaps')
        self.__k_gaps = 0
        self.__k_gap_after_layer = np.zeros((self.__nk - 1,), dtype = bool)
        self.__k_gap_raw_index = np.empty((self.__nk,), dtype = int)
        self.__k_gap_raw_index[0] = 0
        for k in range(self.__nk - 1):
            max_layer_gap = np.max(max_gap_by_layer_and_xyz[k])
            if max_layer_gap > self.__max_z_void:
                self.__k_gap_after_layer[k] = True
                self.__k_gaps += 1
            elif max_layer_gap > 0.0:
                # close void (includes shifting x & y)
                log.debug('closing void below layer (0 based): ' + str(k))
                layer_gap = gap[k] * 0.5
                layer_gap_unmasked = np.where(gap[k].mask, 0.0, layer_gap)
                self.__masked_cp_array[k + 1, :, :, 0, :, :, :] -= layer_gap_unmasked
                self.__masked_cp_array[k, :, :, 1, :, :, :] += layer_gap_unmasked
            self.__k_gap_raw_index[k + 1] = k + self.__k_gaps

    def __close_gaps(self, gap):
        # close voids (includes shifting x & y)
        log.debug('closing voids')
        gap *= 0.5
        gap_unmasked = np.where(gap.mask, 0.0, gap)
        self.__masked_cp_array[1:, :, :, 0, :, :, :] -= gap_unmasked
        self.__masked_cp_array[:-1, :, :, 1, :, :, :] += gap_unmasked

    def __get_k_reduced_cp_array(self):
        log.debug('reducing k extent of corner point array (sharing points vertically)')
        self.__k_reduced_cp_array = ma.masked_array(np.zeros(
            (self.__nk_plus_1, self.__nj, self.__ni, 2, 2, 3)))  # (nk+1+k_gaps, nj, ni, jp, ip, xyz)
        self.__k_reduced_cp_array[0, :, :, :, :, :] = self.__masked_cp_array[0, :, :, 0, :, :, :]
        self.__k_reduced_cp_array[-1, :, :, :, :, :] = self.__masked_cp_array[-1, :, :, 1, :, :, :]
        if self.__k_gaps:
            self.__get_k_reduced_cp_array_kgaps()
        else:
            slice = self.__masked_cp_array[1:, :, :, 0, :, :, :]
            # where cell geometry undefined, if cell above is defined, take data from cell above with kp = 1 and set shared point defined
            self.__k_reduced_cp_array[1:-1, :, :, :, :, :] = np.where(slice.mask, self.__masked_cp_array[:-1, :, :,
                                                                                                         1, :, :, :],
                                                                      slice)

    def __get_k_reduced_cp_array_kgaps(self):
        raw_k = 1
        for k in range(self.__nk - 1):
            # fill reduced array slice(s) for base of layer k and top of layer k + 1
            if self.__k_gap_after_layer[k]:
                self.__k_reduced_cp_array[raw_k, :, :, :, :, :] = self.__masked_cp_array[k, :, :, 1, :, :, :]
                raw_k += 1
                self.__k_reduced_cp_array[raw_k, :, :, :, :, :] = self.__masked_cp_array[k + 1, :, :, 0, :, :, :]
                raw_k += 1
            else:  # take data from either possible cp slice, whichever is defined
                slice = self.__masked_cp_array[k + 1, :, :, 0, :, :, :]
                self.__k_reduced_cp_array[raw_k, :, :, :, :, :] = np.where(slice.mask,
                                                                           self.__masked_cp_array[k, :, :,
                                                                                                  1, :, :, :], slice)
                raw_k += 1
        assert raw_k == self.__nk + self.__k_gaps

    def __get_primary_pillar_ref(self):
        log.debug('creating primary pillar reference neighbourly indices')
        self.__primary_pillar_jip = np.zeros((self.__nj_plus_1, self.__ni_plus_1, 2),
                                             dtype = 'int')  # (nj + 1, ni + 1, jp:ip)
        self.__primary_pillar_jip[-1, :, 0] = 1
        self.__primary_pillar_jip[:, -1, 1] = 1
        for j in range(self.__nj_plus_1):
            for i in range(self.__ni_plus_1):
                if self.__active_mask_2D[j - self.__primary_pillar_jip[j, i, 0],
                                         i - self.__primary_pillar_jip[j, i, 1]]:
                    continue
                if i > 0 and self.__primary_pillar_jip[j, i, 1] == 0 and self.__active_mask_2D[
                        j - self.__primary_pillar_jip[j, i, 0], i - 1]:
                    self.__primary_pillar_jip[j, i, 1] = 1
                    continue
                if j > 0 and self.__primary_pillar_jip[j, i, 0] == 0 and self.__active_mask_2D[
                        j - 1, i - self.__primary_pillar_jip[j, i, 1]]:
                    self.__primary_pillar_jip[j, i, 0] = 1
                    continue
                if i > 0 and j > 0 and self.__primary_pillar_jip[j, i, 0] == 0 and self.__primary_pillar_jip[
                        j, i, 1] == 0 and self.__active_mask_2D[j - 1, i - 1]:
                    self.__primary_pillar_jip[j, i, :] = 1

    def __get_extra_pillar_ref(self):
        self.__extras_count = np.zeros((self.__nj_plus_1, self.__ni_plus_1),
                                       dtype = 'int')  # count (0 to 3) of extras for pillar
        self.__extras_list_index = np.zeros((self.__nj_plus_1, self.__ni_plus_1),
                                            dtype = 'int')  # index in list of 1st extra for pillar
        self.__extras_list = []  # list of (jp, ip)
        self.__extras_use = np.negative(np.ones((self.__nj, self.__ni, 2, 2),
                                                dtype = 'int'))  # (j, i, jp, ip); -1 means use primary
        if self.__split_pillars:
            self.__get_extra_pillar_ref_split()

    def __get_extra_pillar_ref_split(self):
        log.debug('building extra pillar references for split pillars')
        # loop over pillars
        for j in range(self.__nj_plus_1):
            for i in range(self.__ni_plus_1):
                self.__append_extra_pillars(i, j)

        if len(self.__extras_list) == 0:
            self.__split_pillars = False
        log.debug('number of extra pillars: ' + str(len(self.__extras_list)))

    def __append_single_pillar(self, i, j, ip, jp, col_j, p_col_i, p_col_j, primary_ip, primary_jp):
        col_i = i - ip
        if col_i < 0 or col_i >= self.__ni:
            return  # no column this side of pillar in i
        if jp == primary_jp and ip == primary_ip:
            return  # this column is the primary for this pillar
        discrepancy = np.max(
            np.abs(self.__k_reduced_cp_array[:, col_j, col_i, jp, ip, :] -
                   self.__k_reduced_cp_array[:, p_col_j, p_col_i, primary_jp, primary_ip, :]))
        if discrepancy <= self.__split_tolerance:
            return  # data for this column's corner aligns with primary
        for e in range(self.__extras_count[j, i]):
            eli = self.__extras_list_index[j, i] + e
            pillar_j_extra = j - self.__extras_list[eli][0]
            pillar_i_extra = i - self.__extras_list[eli][1]
            discrepancy = np.max(
                np.abs(self.__k_reduced_cp_array[:, col_j, col_i, jp, ip, :] -
                       self.__k_reduced_cp_array[:, pillar_j_extra, pillar_i_extra, self.__extras_list[eli][0],
                                                 self.__extras_list[eli][1], :]))
            if discrepancy <= self.__split_tolerance:  # data for this corner aligns with existing extra
                self.__extras_use[col_j, col_i, jp, ip] = e
                break
        if self.__extras_use[col_j, col_i, jp, ip] >= 0:  # reusing an existing extra for this pillar
            return
        # add this corner as an extra
        if self.__extras_count[j, i] == 0:  # create entry point for this pillar in extras
            self.__extras_list_index[j, i] = len(self.__extras_list)
        self.__extras_list.append((jp, ip))
        self.__extras_use[col_j, col_i, jp, ip] = self.__extras_count[j, i]
        self.__extras_count[j, i] += 1

    def __append_extra_pillars(self, i, j):
        primary_jp = self.__primary_pillar_jip[j, i, 0]
        primary_ip = self.__primary_pillar_jip[j, i, 1]
        p_col_j = j - primary_jp
        p_col_i = i - primary_ip
        # loop over 4 columns surrounding this pillar
        for jp in range(2):
            col_j = j - jp
            if col_j < 0 or col_j >= self.__nj:
                continue  # no column this side of pillar in j
            for ip in range(2):
                self.__append_single_pillar(i, j, ip, jp, col_j, p_col_i, p_col_j, primary_ip, primary_jp)

    def __get_points_array(self):
        log.debug('creating points array as used in resqml format')
        if self.__split_pillars:
            self.__get_points_array_split()
        else:  # unsplit pillars
            self.__points_array = np.zeros((self.__nk_plus_1, self.__nj_plus_1, self.__ni_plus_1, 3))
            for j in range(self.__nj_plus_1):
                for i in range(self.__ni_plus_1):
                    (jp, ip) = self.__primary_pillar_jip[j, i]
                    slice = self.__k_reduced_cp_array[:, j - jp, i - ip, jp, ip, :]
                    self.__points_array[:, j, i, :] = np.where(slice.mask, np.nan,
                                                               slice)  # NaN indicates undefined/invalid geometry

    def __get_points_array_split(self):
        self.__points_array = np.zeros(
            (self.__nk_plus_1, (self.__nj_plus_1 * self.__ni_plus_1) + len(self.__extras_list),
             3))  # note: nk_plus_1 might include k_gaps
        index = 0
        index = self.__get_points_array_split_primary(index)
        index = self.__get_points_array_split_extras(index)

        assert (index == (self.__nj_plus_1 * self.__ni_plus_1) + len(self.__extras_list))

    def __get_points_array_split_primary(self, index):
        # primary pillars
        for pillar_j in range(self.__nj_plus_1):
            for pillar_i in range(self.__ni_plus_1):
                (jp, ip) = self.__primary_pillar_jip[pillar_j, pillar_i]
                slice = self.__k_reduced_cp_array[:, pillar_j - jp, pillar_i - ip, jp, ip, :]
                self.__points_array[:, index, :] = np.where(slice.mask, np.nan,
                                                            slice)  # NaN indicates undefined/invalid geometry
                index += 1
        return index

    def __get_points_array_split_extras(self, index):
        # add extras for split pillars
        for pillar_j in range(self.__nj_plus_1):
            for pillar_i in range(self.__ni_plus_1):
                for e in range(self.__extras_count[pillar_j, pillar_i]):
                    eli = self.__extras_list_index[pillar_j, pillar_i] + e
                    (jp, ip) = self.__extras_list[eli]
                    pillar_j_extra = pillar_j - jp
                    pillar_i_extra = pillar_i - ip
                    slice = self.__k_reduced_cp_array[:, pillar_j_extra, pillar_i_extra, jp, ip, :]
                    self.__points_array[:, index, :] = np.where(slice.mask, np.nan,
                                                                slice)  # NaN indicates undefined/invalid geometry
                    index += 1
        return index

    def __make_basic_grid(self):
        log.debug('initialising grid object')
        self.grid = grr.Grid(self.__model)
        self.grid.grid_representation = 'IjkGrid'
        self.grid.extent_kji = np.array((self.__nk, self.__nj, self.__ni), dtype = 'int')
        self.grid.nk, self.grid.nj, self.grid.ni = self.__nk, self.__nj, self.__ni
        self.grid.k_direction_is_down = True  # assumed direction for corp; todo: determine from geometry and crs z_inc_down flag
        if self.__known_to_be_straight:
            self.grid.pillar_shape = 'straight'
        else:
            self.grid.pillar_shape = 'curved'
        self.grid.has_split_coordinate_lines = self.__split_pillars
        self.grid.k_gaps = self.__k_gaps
        self.grid.k_gap_after_array = self.__k_gap_after_layer
        self.grid.k_raw_index_array = self.__k_gap_raw_index

        self.grid.crs_uuid = self.__crs_uuid
        self.grid.crs_root = self.__model.root_for_uuid(self.__crs_uuid)
        self.__crs = rqc.Crs(self.__model, uuid = self.__crs_uuid)

        # add pillar points array to grid object
        log.debug('attaching points array to grid object')

        self.grid.points_cached = self.__points_array  # NB: reference to points_array, array not copied here

    def __get_split_pillars_lists(self, pillar_i, pillar_j, e):
        self.__split_pillar_indices_list.append((pillar_j * self.__ni_plus_1) + pillar_i)
        use_count = 0
        for jp in range(2):
            j = pillar_j - jp
            if j < 0 or j >= self.__nj:
                continue
            for ip in range(2):
                i = pillar_i - ip
                if i < 0 or i >= self.__ni:
                    continue
                if self.__extras_use[j, i, jp, ip] == e:
                    use_count += 1
                    self.__cols_for_extra_pillar_list.append((j * self.__ni) + i)
        assert (use_count > 0)
        self.__cumulative_length += use_count
        self.__cumulative_length_list.append(self.__cumulative_length)

    def __add_split_arrays_to_grid(self):
        if self.__split_pillars:
            log.debug('adding split pillar arrays to grid object')
            self.__split_pillar_indices_list = []
            self.__cumulative_length_list = []
            self.__cols_for_extra_pillar_list = []
            self.__cumulative_length = 0
            for pillar_j in range(self.__nj_plus_1):
                for pillar_i in range(self.__ni_plus_1):
                    for e in range(self.__extras_count[pillar_j, pillar_i]):
                        self.__get_split_pillars_lists(pillar_i, pillar_j, e)
            log.debug('number of extra pillars: ' + str(len(self.__split_pillar_indices_list)))
            assert (len(self.__cumulative_length_list) == len(self.__split_pillar_indices_list))
            self.grid.split_pillar_indices_cached = np.array(self.__split_pillar_indices_list, dtype = 'int')
            log.debug('number of uses of extra pillars: ' + str(len(self.__cols_for_extra_pillar_list)))
            assert (len(self.__cols_for_extra_pillar_list) == np.count_nonzero(self.__extras_use + 1))
            assert (len(self.__cols_for_extra_pillar_list) == self.__cumulative_length)
            self.grid.cols_for_split_pillars = np.array(self.__cols_for_extra_pillar_list, dtype = 'int')
            assert (len(self.__cumulative_length_list) == len(self.__extras_list))
            self.grid.cols_for_split_pillars_cl = np.array(self.__cumulative_length_list, dtype = 'int')
            self.grid.split_pillars_count = len(self.__extras_list)

    def __set_up_column_to_pillars_mapping(self):
        log.debug('setting up column to pillars mapping')
        base_pillar_count = self.__nj_plus_1 * self.__ni_plus_1
        self.grid.pillars_for_column = np.empty((self.__nj, self.__ni, 2, 2), dtype = 'int')
        for j in range(self.__nj):
            for i in range(self.__ni):
                for jp in range(2):
                    for ip in range(2):
                        if not self.__split_pillars or self.__extras_use[j, i, jp, ip] < 0:  # use primary pillar
                            pillar_index = (j + jp) * self.__ni_plus_1 + i + ip
                        else:
                            eli = self.__extras_list_index[j + jp, i + ip] + self.__extras_use[j, i, jp, ip]
                            pillar_index = base_pillar_count + eli
                        self.grid.pillars_for_column[j, i, jp, ip] = pillar_index

    def __update_grid_geometry_information(self):
        # add cell geometry defined array to model (using active cell mask unless geometry_defined_everywhere is True)
        if self.__geometry_defined_everywhere:
            self.grid.geometry_defined_for_all_cells_cached = True
            self.grid.array_cell_geometry_is_defined = None
        else:
            log.debug('using active cell mask as indicator of defined cell geometry')
            self.grid.array_cell_geometry_is_defined = self.__active_mask.copy(
            )  # a bit harsh: disallows reactivation of cells
            self.grid.geometry_defined_for_all_cells_cached = np.all(self.__active_mask)
        self.grid.geometry_defined_for_all_pillars_cached = True  # following fesapi convention of defining all pillars regardless
        # note: grid.array_pillar_geometry_is_defined not set, as line above should be sufficient

    def __update_grid_handedness(self):
        # set handedness of ijk axes
        if self.__ijk_handedness is None or self.__ijk_handedness == 'auto':
            # work out handedness from sample cell / column axes directions and handedness of crs
            sample_kji0 = tuple(np.array(self.grid.extent_kji) // 2)
            if not self.__geometry_defined_everywhere and not self.grid.array_cell_geometry_is_defined[sample_kji0]:
                where_defined = np.where(
                    np.logical_and(self.grid.array_cell_geometry_is_defined, np.logical_not(self.grid.pinched_out())))
                assert len(where_defined) == 3 and len(where_defined[0]) > 0, 'no extant cell geometries'
                sample_kji0 = (where_defined[0][0], where_defined[1][0], where_defined[2][0])
            sample_cp = self.__cp_array[sample_kji0]
            self.__cell_ijk_lefthanded = (vec.clockwise(sample_cp[0, 0, 0], sample_cp[0, 1, 0], sample_cp[0, 0, 1]) >=
                                          0.0)
            if not self.grid.k_direction_is_down:
                self.__cell_ijk_lefthanded = not self.__cell_ijk_lefthanded
            if self.__crs.is_right_handed_xyz():
                self.__cell_ijk_lefthanded = not self.__cell_ijk_lefthanded
            self.grid.grid_is_right_handed = not self.__cell_ijk_lefthanded
        else:
            assert self.__ijk_handedness in ['left', 'right']
            self.grid.grid_is_right_handed = (self.__ijk_handedness == 'right')

    def create_grid(self):
        """Make the grid"""
        # Find out which cells to treat as nans
        self.__get_treat_as_nan()

        self.__geometry_defined_everywhere = (self.__treat_as_nan is None)

        assert self.__cp_array.ndim == 7
        # Get the grid extents
        self.__get_extents()
        # Generate active and inactive masks
        self.__get_active_inactive_masks()
        # Generate the nan mask
        self.__get_nan_mask()
        # Apply nan and inactive masks
        if self.__cp_nan_mask is not None:
            self.__inactive_mask = np.logical_or(self.__inactive_mask, self.__cp_nan_mask)
            self.__active_mask = np.logical_not(self.__inactive_mask)
        # Generate the masked corner point array
        self.__get_masked_cp_array()

        # Find information on kgaps in the grid
        self.__check_for_kgaps()

        if self.__k_gaps:
            self.__nk_plus_1 += self.__k_gaps
        if self.__k_gap_raw_index is None:
            self.__k_gap_raw_index = np.arange(self.__nk, dtype = int)

        # reduce cp array extent in k
        self.__get_k_reduced_cp_array()

        # create 2D array of active columns (columns where at least one cell is active)
        log.debug('creating 2D array of active columns')
        self.__active_mask_2D = np.any(self.__active_mask, axis = 0)

        # create primary pillar reference indices as one of four column corners around pillar, active column preferred
        self.__get_primary_pillar_ref()

        # build extra pillar references for split pillars
        self.__get_extra_pillar_ref()

        # create points array as used in resqml
        self.__get_points_array()

        # create an empty grid object and fill in some basic info
        self.__make_basic_grid()

        # add split pillar arrays to grid object
        self.__add_split_arrays_to_grid()

        # following is not part of resqml standard but is used by resqml_grid module for speed optimisation
        self.__set_up_column_to_pillars_mapping()

        # add inactive cell mask to grid
        log.debug('setting inactive cell mask')
        self.grid.inactive = self.__inactive_mask.copy()

        # update grid with geometry parameters
        self.__update_grid_geometry_information()

        # tentatively add corner point array to grid object in case it is needed
        log.debug('noting corner point array in grid')
        self.grid.array_corner_points = self.__cp_array

        # update grid handedness
        self.__update_grid_handedness()
